
import os
import time
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import scipy.sparse as sp
import scanpy as sc
import gc
import anndata
import polars as pl
from pathlib import Path
import scipy.io
import zipfile
import decoupler as dc
import shutil

from concurrent.futures import ProcessPoolExecutor
anndata.settings.allow_write_nullable_strings = True


#####---------------------------------------------------------------------
##Creating large merged object from all pituitary datasets
#####---------------------------------------------------------------------

def remove_redundant_cols(adata):
  # Use a lambda function to check if column name contains '0' but is not exactly '0'
  filter_condition = lambda x: '0' in str(x) and str(x) != '0'

  # Apply the filter to both adata.obs and adata.var
  adata.obs = adata.obs.loc[:, ~adata.obs.columns.map(filter_condition)]
  adata.var = adata.var.loc[:, ~adata.var.columns.map(filter_condition)]
  adata.obs = adata.obs.loc[:, ~adata.obs.columns.str.contains('-1')]
  adata.var = adata.var.loc[:, ~adata.var.columns.str.contains('-1')]

  #rename obs from 1-len
  adata.obs.index = [f"{i+1}" for i in range(len(adata.obs))]

  #turn to string so saving won't throw error
  adata.obs.columns = adata.obs.columns.astype(str)
  adata.var.columns = adata.var.columns.astype(str)

  for col in adata.obs.columns:
    adata.obs[col] = adata.obs[col].astype(str)

  #remove obsm and layers to make object more lightweight

  adata.obsm = {}
  adata.layers = {}

  return adata


def remove_low_quality_clusters(adata, scvi_labels_to_keep, SRA_ID, df, ignore_core=False):

  author = df[df["SRA_ID"]==SRA_ID]["Author"].values[0]
  GEO = df[df["SRA_ID"]==SRA_ID]["GEO"].values[0]

  if ignore_core: #if not core sample, therefore not in scvi labels to keep, still keep the data by skipping filtering
    print(f"Ignoring core filtering for {SRA_ID}")
    return adata
  
  elif author == "Rebboah et al. (2025)":

    adata.obs["barcode"] = [GEO + ":" + adata.obs["bc"][i] + "_" + adata.obs["subpool"][i] + "-1"  for i in range(len(adata.obs))]
    shape1 = adata.shape
    adata = adata[adata.obs["barcode"].isin(scvi_labels_to_keep[scvi_labels_to_keep["SRA_ID"]==SRA_ID]["barcode"].values)]
    shape2 = adata.shape
    print(f"Removed {shape1[0]-shape2[0]} cells")

  elif author == "Sochodolsky et al. (2026)":
     
    adata.obs["barcode"] = [SRA_ID+":" + adata.obs['0'][i] + "-1" for i in range(len(adata.obs))]
    shape1 = adata.shape
    adata = adata[adata.obs["barcode"].isin(scvi_labels_to_keep[scvi_labels_to_keep["SRA_ID"]==SRA_ID]["barcode"].values)]
    shape2 = adata.shape
    print(f"Removed {shape1[0]-shape2[0]} cells")

  else:
    adata.obs["barcode"] = [GEO+":" + adata.obs['0'][i] + "-1" for i in range(len(adata.obs))]
    shape1 = adata.shape
    adata = adata[adata.obs["barcode"].isin(scvi_labels_to_keep[scvi_labels_to_keep["SRA_ID"]==SRA_ID]["barcode"].values)]
    shape2 = adata.shape
    print(f"Removed {shape1[0]-shape2[0]} cells")

  return adata


def generate_merged_object(nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                           remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                           back_up_remaining_path="/analysis/adata_assigned.h5ad",
                           atlas_source_path='/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
                           output_path="/content/drive/MyDrive/pituitary_atlas/Unstructured/wt_adata_merged_post_scvi_0902.h5ad",
                           filtering_based_on_barcodes = False,
                           scvi_labels_to_keep = None,
                           modality = ['sc','sn','multi_rna'],
                           normal = [1],
                           n_cells_threshold = 300,
                           ID_col='SRA_ID',
                           include_new_samples = False #samples that are Core == 2
                           ):
    
    gc.collect()
    df = pd.read_excel(atlas_source_path)
    pituitary_atlas = df.copy()
    df = df[df['Modality'].isin(modality)]
    df = df[df["Normal"].isin(normal)]
    df = df[df["n_cells"]>n_cells_threshold]
    df.reset_index(drop=True, inplace=True)
    adata_merged_flag = False
    for i in range(len(df)-1,-1,-1):
      try:
          
          ID = df[ID_col][i]
          core_value = df[df[ID_col]==ID]["Core"].values[0]
          print(ID)
          try:
            adata = sc.read(f"{nac_path}{ID}{remaining_path}")
          except:
            print("Falling back to adata_assigned")
            adata = sc.read(f"{nac_path}{ID}{back_up_remaining_path}")

          #ensure X is sparse!
          if not sp.issparse(adata.X):
              adata.X = sp.csr_matrix(adata.X)
          print("Loaded adata shape:", adata.shape)
          adata.obs[ID_col] = ID
          print("Added SRA_ID to obs")
          adata.obs["GEO"] = df[df[ID_col]==ID]["GEO"].values[0]
          print("Added GEO to obs")
          adata.obs["assay"] = df[df[ID_col]==ID]["Modality"].values[0]
          print("Added Modality to obs")

          adata.obs["orig_index"] = adata.obs.index.astype(str)
          core_action = False
          if include_new_samples and core_value == 2:
            core_action = True #true, therefore ignore filter and keep all cells
             
          if filtering_based_on_barcodes:
            adata = remove_low_quality_clusters(adata, scvi_labels_to_keep, ID, df,ignore_core=core_action)
          
          if not include_new_samples and core_value == 2:
            print(f"Not including non-core samples, removing {ID}")
            continue
            

          if ("1" in adata.var_names and "2" in adata.var_names and "3" in adata.var_names) or (1 in adata.var_names and 2 in adata.var_names and 3 in adata.var_names.astype(str)):
            print("The problematic sample is:", ID)
            break

          if adata_merged_flag == False:
              adata_merged = adata.copy()
              adata_merged_flag = True
          else:
              adata_merged = anndata.concat([adata_merged, adata], join="outer", uns_merge=None)
              adata_merged.obs_names_make_unique()
          print(adata_merged.shape)
          del adata

          #sleep 5 seconds
          gc.collect()
          time.sleep(2)
          

      #except print error
      except Exception as e:
          print(f"Error processing {ID}: {str(e)}")
          continue

    print(adata_merged.shape)
    remove_redundant_cols(adata_merged).write(output_path, compression='gzip')


#####---------------------------------------------------------------------
##Generate a matrix of cell counts per cell type across datasets.
#####---------------------------------------------------------------------

def generate_cell_count_matrix(df,
                            nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                           remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                           back_up_remaining_path="/analysis/adata_assigned.h5ad",
                           filtering_based_on_barcodes = False,
                           scvi_labels_to_keep = None,
                           min_cells=50,
                           cell_type_col='new_cell_type',
                           cell_type_col_backup='assignments',
                           ID_col='SRA_ID',
                           include_new_samples = False

                             ):


    """
    Generate a matrix of cell counts per cell type across datasets.

    """

    

    all_cell_types = set()
    dataset_counts = {}
    total_counts = {}

    # First pass: collect all cell types and counts
    print("First pass: collecting cell types and counts...")
    for i in range(len(df)):
        ID = df[ID_col][i]
        print(ID)
        core_value = df[df[ID_col]==ID]["Core"].values[0]
        core_action = False
        if include_new_samples and core_value == 2:
          core_action = True

        try:
            print(f"\nProcessing {ID}...")
            try:
              adata = sc.read(f"{nac_path}{ID}{remaining_path}")
              if filtering_based_on_barcodes:
                adata = remove_low_quality_clusters(adata, scvi_labels_to_keep, ID, df, ignore_core=core_action)
            except:
              print("Falling back to adata_assigned")
              adata = sc.read(f"{nac_path}{ID}{back_up_remaining_path}")
              if filtering_based_on_barcodes:
                adata = remove_low_quality_clusters(adata, scvi_labels_to_keep, ID, df, ignore_core=core_action)

            try:
              cell_type_counts = adata.obs[cell_type_col].astype(str).value_counts()
            except:
              cell_type_counts = adata.obs[cell_type_col_backup].astype(str).value_counts()
            valid_types = cell_type_counts[cell_type_counts >= min_cells]

            # Store counts and update cell types set
            dataset_counts[ID] = valid_types
            all_cell_types.update(valid_types.index)
            total_counts[ID] = len(adata)

            print(f"Found {len(valid_types)} cell types with ≥{min_cells} cells")
            print(f"Total cells: {total_counts[ID]}")

        except Exception as e:
            print(f"Error processing {ID}: {str(e)}")
            continue

    # Sort cell types for consistent column order
    all_cell_types = sorted(list(all_cell_types))
    print(f"\nTotal unique cell types across all datasets: {len(all_cell_types)}")

    # Create matrix with datasets as rows and cell types as columns
    matrix_data = []
    for ID in dataset_counts.keys():
        row_data = {'dataset': ID}
        counts = dataset_counts[ID]

        # Fill in counts for each cell type
        for cell_type in all_cell_types:
            row_data[cell_type] = int(counts.get(cell_type, 0))

        matrix_data.append(row_data)

    # Convert to polars DataFrame
    cell_count_matrix = pl.DataFrame(matrix_data)

    return cell_count_matrix, total_counts


def save_count_matrix(matrix, total_counts, output_dir="/content/results"):
    """
    Save the cell count matrix and related information.
    """
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    # Save main count matrix
    matrix.write_csv(f"{output_dir}/cell_type_counts.csv")

    # Save total counts
    total_counts_df = pl.DataFrame({
        'dataset': list(total_counts.keys()),
        'total_cells': list(total_counts.values())
    })
    total_counts_df.write_csv(f"{output_dir}/total_cell_counts.csv")



def dataset_ct_matrix(atlas_source_path = '/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
                      output_dir="output_directory",
                      Modality = ['sc','sn','multi_rna'],
                      n_cells_threshold = 300,

                      nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                      remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                      back_up_remaining_path="/analysis/adata_assigned.h5ad",
                      filtering_based_on_barcodes = False,
                      scvi_labels_to_keep = None,
                      min_cells=50,
                      cell_type_col='new_cell_type',
                      cell_type_col_backup='assignments',
                      ID_col='SRA_ID',
                      include_new_samples = False
                      ):
  #to display users at the start how many cell of each cell type we have
  import pandas as pd
  df = pd.read_excel(atlas_source_path)
  df = df[df['Modality'].isin(Modality)]
  df = df[df["n_cells"]>n_cells_threshold]
  df.reset_index(drop=True, inplace=True)
  #reset
  df = df.reset_index(drop=True)
  count_matrix, totals = generate_cell_count_matrix(
                              df,
                              nac_path=nac_path,
                              remaining_path=remaining_path,
                              back_up_remaining_path=back_up_remaining_path,
                              filtering_based_on_barcodes = filtering_based_on_barcodes,
                              scvi_labels_to_keep = scvi_labels_to_keep,
                              min_cells=min_cells,
                              cell_type_col=cell_type_col,
                              cell_type_col_backup=cell_type_col_backup,
                              ID_col=ID_col,
                              include_new_samples = include_new_samples
                              )
  save_count_matrix(count_matrix, totals, output_dir)

#---------------------------------------------------------------------
# Generate values for dotplot construction
#---------------------------------------------------------------------


def get_expression_metrics(adata,
                           min_abundance=50,
                           cell_type_col='new_cell_type',
                           cell_type_col_backup='assignments',
                            min_expr=0,
                            name=''):


    """Calculate expression metrics with genes as columns"""
    # Get cell types and filter by minimum cell count (50)
    try:
      cell_type_counts = adata.obs[cell_type_col].astype(str).value_counts()
    except:
      cell_type_counts = adata.obs[cell_type_col_backup].astype(str).value_counts()
    valid_cell_types = cell_type_counts[cell_type_counts >= min_abundance].index.values
    print(f"Cell types with >={min_abundance} cells: {len(valid_cell_types)} out of {len(cell_type_counts)}")
    genes = adata.var_names
    print(f"Number of genes: {len(genes)}")

    # Initialize result arrays
    n_cell_types = len(valid_cell_types)
    n_genes = len(genes)
    means = np.zeros((n_cell_types, n_genes))
    fracs = np.zeros((n_cell_types, n_genes))

    # Convert sparse matrix to dense once
    X = adata.X.toarray() if sp.issparse(adata.X) else np.array(adata.X)
    print(f"Matrix shape: {X.shape}")

    # Calculate metrics for each valid cell type
    for i, ct in enumerate(valid_cell_types):
        try:
          mask = adata.obs[cell_type_col].astype(str) == ct
          ct_expr = X[mask]
        except:
          mask = adata.obs[cell_type_col_backup].astype(str) == ct
          ct_expr = X[mask]
        print(f"  Cell type {ct}: {sum(mask)} cells")

        # Vectorized calculations
        means[i] = np.mean(ct_expr, axis=0)
        fracs[i] = np.mean(ct_expr > min_expr, axis=0)

    # Create DataFrames with SRA_ID prefixed cell types
    prefixed_cell_types = [f"{name}_{ct}" if name else ct for ct in valid_cell_types]

    # Create dictionary for DataFrame construction
    means_dict = {'cell_type': prefixed_cell_types}
    fracs_dict = {'cell_type': prefixed_cell_types}

    for j, gene in enumerate(genes):
        means_dict[gene] = means[:, j]  # Use all rows for this gene column
        fracs_dict[gene] = fracs[:, j]  # Use all rows for this gene column

    means_df = pl.DataFrame(means_dict)
    fracs_df = pl.DataFrame(fracs_dict)

    return means_df, fracs_df




def process_all_datasets(df,
                         nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                           remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                           back_up_remaining_path="/analysis/adata_assigned.h5ad",
                           filtering_based_on_barcodes = False,
                           scvi_labels_to_keep = None,
                           ID_col='SRA_ID',
                           min_abundance=50,
                           cell_type_col='new_cell_type',
                           cell_type_col_backup='assignments',
                            min_expr=0,
                            include_new_samples = False
                           ):
    
    """Process datasets with genes as columns"""
    all_matrix1s = []
    all_matrix2s = []
    all_genes = set()

    # First pass: collect all genes
    print("First pass: collecting all genes...")
    for i in range(len(df)):
        ID = df[ID_col][i]
        print(f"\nProcessing {ID}...")
        try:
            try:
              adata = sc.read(f"{nac_path}{ID}{remaining_path}")
            except:
              print("Falling back to adata_assigned")
              adata = sc.read(f"{nac_path}{ID}{back_up_remaining_path}")
            all_genes.update(adata.var_names)
        except Exception as e:
            print(f"Error reading genes from {ID}: {str(e)}")
            continue

    all_genes = sorted(list(all_genes))
    print(f"Total unique genes across all datasets: {len(all_genes)}")

    # Second pass: process datasets and align genes
    print("\nSecond pass: processing datasets...")
    for i in range(len(df)):
        ID = df[ID_col][i]
        core_value = df[df[ID_col]==ID]["Core"].values[0]
        core_action = False
        if include_new_samples and core_value == 2:
          core_action = True
        try:
            print(f"\nProcessing {ID}...")
            try:
              adata = sc.read(f"{nac_path}{ID}{remaining_path}")
              if filtering_based_on_barcodes:
                adata = remove_low_quality_clusters(adata, scvi_labels_to_keep, ID, df, ignore_core=core_action)
              
            except:
              print("Falling back to adata_assigned")
              adata = sc.read(f"{nac_path}{ID}{back_up_remaining_path}")
              if filtering_based_on_barcodes:
                adata = remove_low_quality_clusters(adata, scvi_labels_to_keep, ID, df, ignore_core=core_action)


            if ("1" in adata.var_names) or ("10" in adata.var_names) or (10 in adata.var_names) or (1 in adata.var_names):
              #throw error and break
              raise Exception("1 or 10 in var_names")
              break

            #total norm and log1p
            sc.pp.normalize_total(adata, target_sum=1e4)
            sc.pp.log1p(adata)

            # Get metrics with genes as columns
            matrix_1, matrix_2 = get_expression_metrics(adata, 
                                                        min_abundance=min_abundance,
                                                        cell_type_col= cell_type_col,
                                                        cell_type_col_backup=cell_type_col_backup,
                                                        min_expr=min_expr,
                                                        name=ID)
            print(f"Initial matrix shapes - M1: {matrix_1.shape}, M2: {matrix_2.shape}")

            # Add missing genes with zeros
            missing_genes = set(all_genes) - set(matrix_1.columns[1:])  # Skip cell_type column
            print(f"Adding {len(missing_genes)} missing genes")
            if missing_genes:
                new_cols = {gene: pl.lit(0.0) for gene in missing_genes}
                matrix_1 = matrix_1.with_columns(**new_cols)
                matrix_2 = matrix_2.with_columns(**new_cols)

            # Ensure columns are in the same order
            col_order = ['cell_type'] + sorted(all_genes)
            matrix_1 = matrix_1.select(col_order)
            matrix_2 = matrix_2.select(col_order)

            all_matrix1s.append(matrix_1)
            all_matrix2s.append(matrix_2)

            print(f"Completed {ID}, shape: {matrix_1.shape}")

        except Exception as e:
            print(f"Error processing {ID}: {str(e)}")
            continue

    # Combine matrices by concatenating rows
    if all_matrix1s and all_matrix2s:
        final_matrix1 = pl.concat(all_matrix1s, how="vertical")
        final_matrix2 = pl.concat(all_matrix2s, how="vertical")
        return final_matrix1, final_matrix2
    else:
        raise ValueError("No datasets were successfully processed")



def save_results(final_matrix1, final_matrix2, output_dir="results"):
    """Save results with genes as columns"""
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    for idx, (matrix, name) in enumerate([(final_matrix1, "matrix1"), (final_matrix2, "matrix2")]):
        print(f"Saving {name}")

        # Save cell types list
        cell_types = matrix.select("cell_type").to_series().to_list()
        with open(f"{output_dir}/{name}_rows.tsv", 'w') as f:
            f.write('\n'.join(cell_types))

        # Save gene names (columns)
        gene_cols = [col for col in matrix.columns if col != "cell_type"]
        with open(f"{output_dir}/{name}_genes.tsv", 'w') as f:
            f.write('\n'.join(gene_cols))

        # Save matrix data
        data_matrix = matrix.drop("cell_type").to_numpy()
        sparse_matrix = sp.csr_matrix(data_matrix)
        scipy.io.mmwrite(f"{output_dir}/{name}.mtx", sparse_matrix)


def dotplot_matrix(atlas_source_path = '/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
                  output_dir="results",
                  Modality = ['sc','sn','multi_rna'],
                  n_cells_threshold = 300,
                  nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                           remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                           back_up_remaining_path="/analysis/adata_assigned.h5ad",
                           filtering_based_on_barcodes = False,
                           scvi_labels_to_keep = None,
                           ID_col='SRA_ID',
                           min_abundance=50,
                           cell_type_col='new_cell_type',
                           cell_type_col_backup='assignments',
                            min_expr=0,
                            include_new_samples = False
                  ):

  df = pd.read_excel(atlas_source_path)
  pituitary_atlas = df.copy()
  df = df[df['Modality'].isin(Modality)]
  df = df[df["n_cells"]>n_cells_threshold]
  df.reset_index(drop=True, inplace=True)

  final_matrix1, final_matrix2 = process_all_datasets(df,
                              nac_path=nac_path,
                           remaining_path=remaining_path,
                           back_up_remaining_path=back_up_remaining_path,
                           filtering_based_on_barcodes = filtering_based_on_barcodes,
                           scvi_labels_to_keep = scvi_labels_to_keep,
                           ID_col=ID_col,
                           min_abundance=min_abundance,
                           cell_type_col=cell_type_col,
                           cell_type_col_backup=cell_type_col_backup,
                            min_expr=min_expr,
                            include_new_samples = include_new_samples

                              )
  save_results(final_matrix1, final_matrix2, output_dir=output_dir)



#---------------------------------------------------------------------
# Generate values for cell type proportion analysis
#---------------------------------------------------------------------

def get_cell_type_abundances(adata,
                             cell_type_col='new_cell_type',
                              cell_type_col_backup='assignments'
                              ):
    """Calculate cell type abundances as percentages"""
    try:
      cell_types = adata.obs[cell_type_col].astype(str).value_counts()
    except:
      cell_types = adata.obs[cell_type_col_backup].astype(str).value_counts()
    total_cells = len(adata)

    # Calculate percentages
    percentages = (cell_types / total_cells) * 100
    return percentages.index.values, percentages.values

def process_datasets_abundances(df,
                              nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                            remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                            back_up_remaining_path="/analysis/adata_assigned.h5ad",
                            scvi_labels_to_keep = None,
                            ID_col='SRA_ID',
                            cell_type_col='new_cell_type',
                            cell_type_col_backup='assignments',
                            filtering_based_on_barcodes = False,
                            include_new_samples = False
                            ):
    """Process datasets and create abundance matrix"""
    all_cell_types = set()

    # First pass: collect all cell types
    print("First pass: collecting all cell types...")
    for i in range(len(df)):
        ID = df[ID_col][i]
        try:
            try:
              adata = sc.read(f"{nac_path}{ID}{remaining_path}")
            except:
              print("Falling back to adata_assigned")
              adata = sc.read(f"{nac_path}{ID}{back_up_remaining_path}")
            cell_types, _ = get_cell_type_abundances(adata,
                                                      cell_type_col=cell_type_col,
                                                      cell_type_col_backup=cell_type_col_backup)
            all_cell_types.update(cell_types)
        except Exception as e:
            print(f"Error reading {ID}: {str(e)}")
            continue

    all_cell_types = sorted(list(all_cell_types))
    n_cell_types = len(all_cell_types)
    print(f"Total unique cell types: {n_cell_types}")

    # Create cell type to index mapping
    cell_type_to_idx = {ct: i for i, ct in enumerate(all_cell_types)}

    # Second pass: fill abundance matrix
    abundance_matrix = np.zeros((n_cell_types, len(df)))
    valid_samples = []

    print("\nSecond pass: calculating abundances...")
    for i in range(len(df)):
        ID = df[ID_col][i]
        core_value = df[df[ID_col]==ID]["Core"].values[0]
        core_action = False
        if include_new_samples and core_value == 2:
          core_action = True
        try:
            print(f"Processing {ID}...")
            try:
              adata = sc.read(f"{nac_path}{ID}{remaining_path}")
              if filtering_based_on_barcodes:
                adata = remove_low_quality_clusters(adata, scvi_labels_to_keep, ID, df, ignore_core=core_action)

            except:
              print("Falling back to adata_assigned")
              adata = sc.read(f"{nac_path}{ID}{back_up_remaining_path}")
              if filtering_based_on_barcodes:
                adata = remove_low_quality_clusters(adata, scvi_labels_to_keep, ID, df, ignore_core=core_action)

            cell_types, percentages = get_cell_type_abundances(adata,
                                                              cell_type_col=cell_type_col,
                                                              cell_type_col_backup=cell_type_col_backup)
            # Fill abundance matrix
            for ct, pct in zip(cell_types, percentages):
                abundance_matrix[cell_type_to_idx[ct], i] = pct

            valid_samples.append(ID)
            print(f"Completed {ID}")

        except Exception as e:
            print(f"Error processing {ID}: {str(e)}")
            continue

    return abundance_matrix, all_cell_types, valid_samples


def save_abundance_results(abundance_matrix, cell_types, samples, output_dir="abundance_results"):
    """Save abundance results in MTX format"""
    Path(output_dir).mkdir(parents=True, exist_ok=True)

    # Save cell types (rows)
    with open(f"{output_dir}/abundance_rows.tsv", 'w') as f:
        f.write('\n'.join(cell_types))

    # Save sample IDs (columns)
    with open(f"{output_dir}/abundance_cols.tsv", 'w') as f:
        f.write('\n'.join(samples))

    # Save abundance matrix
    sparse_matrix = sp.csr_matrix(abundance_matrix)
    scipy.io.mmwrite(f"{output_dir}/abundance.mtx", sparse_matrix)


def cell_type_proportions(atlas_source_path = '/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
                         Modality = ['sc','sn','multi_rna'],
                         n_cells_threshold = 300,
                         nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                            remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                            back_up_remaining_path="/analysis/adata_assigned.h5ad",
                            scvi_labels_to_keep = None,
                            ID_col='SRA_ID',
                            cell_type_col='new_cell_type',
                            cell_type_col_backup='assignments',
                            filtering_based_on_barcodes = False,
                          output_dir="abundance_results",
                          include_new_samples = False
                         ):

  df = pd.read_excel(atlas_source_path)
  df = df[df['Modality'].isin(Modality)]
  df = df[df["n_cells"]>n_cells_threshold]
  df.reset_index(drop=True, inplace=True)
  #reset
  df = df.reset_index(drop=True)
  # Usage
  matrix, cell_types, samples = process_datasets_abundances(df,
                              nac_path=nac_path,
                            remaining_path=remaining_path,
                            back_up_remaining_path=back_up_remaining_path,
                            scvi_labels_to_keep = scvi_labels_to_keep,
                            ID_col=ID_col,
                            cell_type_col=cell_type_col,
                            cell_type_col_backup=cell_type_col_backup,
                            filtering_based_on_barcodes = filtering_based_on_barcodes,
                            include_new_samples = include_new_samples

                              )
  save_abundance_results(matrix, cell_types, samples, output_dir=output_dir)


#---------------------------------------------------------------------
# Processing adatas for epitome h5 files
#---------------------------------------------------------------------


def simplify_and_process_adatas(df,
                                output_dir="/content/drive/MyDrive/pituitary_atlas/Unstructured/epitome_h5_files_rna",
                                nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                                remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                                back_up_remaining_path="/analysis/adata_assigned.h5ad"
                                ):
    
    """Simplify and process adatas, then save as h5ad files."""

    for i in range(len(df)):
        SRA_ID = df["SRA_ID"][i]

        #if output_path = f"{output_dir}/{SRA_ID}_processed.h5ad" exists continue
        if Path(f"{output_dir}/{SRA_ID}_processed.h5ad").exists():
            print(f"Skipping {SRA_ID} as it has already been processed")
            continue
        try:
            print(f"\nProcessing {SRA_ID}...")
            try:
              adata = sc.read(f"{nac_path}/{SRA_ID}{remaining_path}")
            except:
              print("Falling back to adata_assigned")
              adata = sc.read(f"{nac_path}/{SRA_ID}{back_up_remaining_path}")

            # Simplify var and obs
            adata.var = adata.var.iloc[:, 0:0]  # Keep only index
            try:
              adata.obs = adata.obs[["new_cell_type"]]
            except:
              adata.obs = adata.obs[["assignments"]]
              #rename to "new_cell_type"
              adata.obs.rename(columns={"assignments": "new_cell_type"}, inplace=True)
            #sc.pp.filter_genes(adata, min_cells=20)
            # Processing pipeline

            adata.layers["counts"] = adata.X.copy()

            sc.pp.normalize_total(adata, target_sum=1e4)
            sc.pp.log1p(adata)
            adata.layers["log1p"] = adata.X.copy()
            sc.pp.highly_variable_genes(adata, n_top_genes=2000)
            sc.pp.scale(adata, max_value=10)
            sc.tl.pca(adata, svd_solver='arpack')
            sc.pp.neighbors(adata, n_neighbors=25, n_pcs=42)
            sc.tl.leiden(adata, flavor='igraph')
            sc.tl.umap(adata)

            adata.X = adata.layers["log1p"].copy()

            sc.pl.umap(adata, color=['new_cell_type'])

            #make sure X is a sparse matrix
            if not sp.issparse(adata.X):
                adata.X = sp.csr_matrix(adata.X)

            #repeat for all layers
            for layer in adata.layers:
                if not sp.issparse(adata.layers[layer]):
                    adata.layers[layer] = sp.csr_matrix(adata.layers[layer])
            # Save processed adata
            output_path = f"{output_dir}/{SRA_ID}_processed.h5ad"
            adata.write_h5ad(output_path, compression='gzip')
            print(f"Saved to {output_path}")

        except Exception as e:
            print(f"Error processing {SRA_ID}: {str(e)}")
            continue


#wrap function to call save_all_h5s
def save_all_h5s(atlas_source_path = '/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
                  Modality = ['sc','sn','multi_rna'],
                  n_cells_threshold = 300,
                  output_dir="/content/drive/MyDrive/pituitary_atlas/Unstructured/epitome_h5_files_rna",
                                nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                                remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                                back_up_remaining_path="/analysis/adata_assigned.h5ad"

                  ):
   
  df = pd.read_excel(atlas_source_path)
  pituitary_atlas = df.copy()
  df = df[df['Modality'].isin(Modality)]
  df = df[df["n_cells"]>n_cells_threshold]
  df.reset_index(drop=True, inplace=True)
  #reset
  df = df.reset_index(drop=True)
  df
  # Usage
  simplify_and_process_adatas(df,
                              output_dir=output_dir,
                              nac_path=nac_path,
                              remaining_path=remaining_path,
                              back_up_remaining_path=back_up_remaining_path
                              )


#function to bundle up all h5 files into zip files of 50 each
def bundle_up_all_h5s(source_folder='/content/drive/MyDrive/pituitary_atlas/Unstructured/epitome_h5_files_rna/',
                      base_output_zip='/content/drive/MyDrive/pituitary_atlas/Unstructured/epitome_h5_files_rna/epitome_grouped_h5_files_0828'):
  """Bundles h5 files into zip files of specified chunk size."""
   
  # Get the list of all files in the folder
  all_files = [f for f in os.listdir(source_folder) if os.path.isfile(os.path.join(source_folder, f))]
  len(all_files)

  # Group files into chunks of 50
  chunk_size = 15
  file_chunks = [all_files[i:i + chunk_size] for i in range(0, len(all_files), chunk_size)]

  # Create a zip file for each chunk of files
  for idx, chunk in enumerate(file_chunks):
      zip_filename = f'{base_output_zip}_group_{idx + 1}.zip'
      zip_filepath = os.path.join(source_folder, zip_filename)

      with zipfile.ZipFile(zip_filepath, 'w', zipfile.ZIP_DEFLATED) as zipf:
          for file in chunk:
              file_path = os.path.join(source_folder, file)
              zipf.write(file_path, os.path.basename(file_path))  # Write file to the zip archive

      print(f'Created zip file: {zip_filename}')



def zip_summary_pdfs(atlas_source_path = '/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
                     Modality = ['sc','sn','multi_rna'],
                     n_cells_threshold = 300,
                     output_dir="/content/drive/MyDrive/pituitary_atlas/Unstructured/epitome_pdf_files",
                     output_zip_name="summary_pdfs.zip",
                     pdf_path_prefix="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                     pdf_path_suffix="/analysis/summary.pdf"
                     ):
    """Zips summary PDFs for all SRA IDs in the DataFrame, using SRA_ID in filename."""
    df = pd.read_excel(atlas_source_path)
    pituitary_atlas = df.copy()
    df = df[df['Modality'].isin(Modality)]
    df = df[df["n_cells"]>n_cells_threshold]
    df.reset_index(drop=True, inplace=True)

    Path(output_dir).mkdir(parents=True, exist_ok=True)
    output_zip_path = os.path.join(output_dir, output_zip_name)

    with zipfile.ZipFile(output_zip_path, 'w') as zipf:
        for sra_id in df["SRA_ID"]:
            pdf_path = f"{pdf_path_prefix}{sra_id}{pdf_path_suffix}"
            if os.path.exists(pdf_path):
                # Create new filename with SRA_ID
                new_filename = f"summary_{sra_id}.pdf"
                zipf.write(pdf_path, arcname=new_filename)
                print(f"Added {pdf_path} as {new_filename} to zip")
            else:
                print(f"PDF not found: {pdf_path}")

    print(f"Zip file created at: {output_zip_path}")


import pandas as pd
import scanpy as sc
import os
import gc
import zipfile
from concurrent.futures import ProcessPoolExecutor

# --- 1. Define OUTSIDE the function ---
def save_gene_batch_parquet(gene_list, sparse_X, obs_names, var_names):
    """
    Worker function: Receives only the essential data pieces.
    This avoids NameErrors and keeps RAM usage predictable.
    """
    # Find positions of requested genes
    gene_indices = [var_names.get_loc(g) for g in gene_list]
    
    # Extract columns and convert to dense only for this small batch
    X_batch = sparse_X[:, gene_indices].toarray()

    batch_df = pd.DataFrame(
        data=X_batch,
        index=obs_names,
        columns=gene_list
    )

    file_paths = []
    for gene in gene_list:
        file_path = f"genes_parquet/{gene}.parquet"
        # Using zstd for speed and efficiency
        batch_df[[gene]].to_parquet(file_path, compression="zstd")
        file_paths.append(file_path)

    del batch_df, X_batch
    gc.collect()
    return file_paths

# --- 2. The Main Function ---
def generate_atlas_umap(
    atlas_path='/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
    dataset_with_coords="/content/drive/MyDrive/pituitary_atlas/Unstructured/scanvi_all_cells_0828.h5ad",
    wt_dataset="/content/drive/MyDrive/pituitary_atlas/Unstructured/wt_adata_merged_post_scvi_0902.h5ad",
    mut_dataset="/content/drive/MyDrive/pituitary_atlas/Unstructured/mut_adata_merged_post_scvi_0902.h5ad",
    output_total_counts_csv="/content/drive/MyDrive/pituitary_atlas/Unstructured/total_counts.csv",
    zip_output_path="/content/drive/MyDrive/pituitary_atlas/Unstructured/adata_export_large_umap_1015.zip",
):
    # Load and process data
    adata = sc.read_h5ad(dataset_with_coords)
    adata_wt = sc.read_h5ad(wt_dataset)
    adata_mut = sc.read_h5ad(mut_dataset)

    # 1. Create IDs
    for d in [adata_wt, adata_mut, adata]:
        d.obs["cell_id"] = d.obs["SRA_ID"].astype(str) + "_" + d.obs["0"].astype(str)
    
    # Handle NaNs for parse samples
    nans = adata.obs["0"] == "nan"
    adata.obs.loc[nans, "cell_id"] = adata.obs.loc[nans, "SRA_ID"].astype(str) + "_" + \
                                     adata.obs.loc[nans, "bc"].astype(str) + "_" + \
                                     adata.obs.loc[nans, "subpool"].astype(str)

    # 2. Merge data using the new concat method
    merged_adata = sc.concat([adata_wt, adata_mut], label="batch")
    merged_adata.obs_names_make_unique() # STOPS THE WARNING AND RAM SPIKES
    merged_adata = merged_adata.copy()

    # 3. Map UMAPs
    mapping_umap1 = dict(zip(adata.obs["cell_id"], adata.obs["UMAP1"]))
    mapping_umap2 = dict(zip(adata.obs["cell_id"], adata.obs["UMAP2"]))
    merged_adata.obs["UMAP1"] = merged_adata.obs["cell_id"].map(mapping_umap1)
    merged_adata.obs["UMAP2"] = merged_adata.obs["cell_id"].map(mapping_umap2)
    merged_adata = merged_adata[~merged_adata.obs["UMAP1"].isna()].copy()

    # 4. Total Counts
    merged_adata.obs["ncounts"] = merged_adata.X.sum(axis=1).A1
    merged_adata.obs["ncounts"].to_csv(output_total_counts_csv)

    # 5. Metadata merge (your atlas logic)
    pituitary_atlas = pd.read_excel(atlas_path).drop_duplicates(subset=["SRA_ID"])
    pituitary_atlas = pituitary_atlas[["Comp_sex","10X version","Modality","Normal","SRA_ID"]]
    
    for col in ["Comp_sex","10X version","Modality","Normal","Sex"]:
        if col in merged_adata.obs.columns:
            merged_adata.obs.drop(col, axis=1, inplace=True)
            
    merged_adata.obs = merged_adata.obs.merge(pituitary_atlas, on="SRA_ID", how="left")

    # 6. Prepare for Export
    if hasattr(merged_adata.X, 'format') and merged_adata.X.format != 'csc':
        merged_adata.X = merged_adata.X.tocsc()

    obs_path = "obs.parquet"
    merged_adata.obs.to_parquet(obs_path, compression="gzip")
    merged_adata.obs[["UMAP1", "UMAP2"]].to_parquet("umap.parquet", compression="gzip")

    # 7. PARALLEL GENE EXPORT
    os.makedirs("genes_parquet", exist_ok=True)
    
    # We pass the sparse matrix and names ONLY. 
    # This uses a fraction of the RAM compared to passing the whole adata object.
    sparse_X = merged_adata.X
    obs_names = merged_adata.obs_names
    var_names = merged_adata.var_names
    
    gene_names = var_names.to_list()
    BATCH_SIZE = 100
    gene_batches = [gene_names[i:i + BATCH_SIZE] for i in range(0, len(gene_names), BATCH_SIZE)]

    # Limit workers in Colab to prevent OOM
    MAX_WORKERS = 2 

    print(f"Starting parallel export of {len(gene_names)} genes...")
    with ProcessPoolExecutor(max_workers=MAX_WORKERS) as executor:
        # We use a list comprehension to pass the specific objects to each task
        futures = [executor.submit(save_gene_batch_parquet, batch, sparse_X, obs_names, var_names) 
                   for batch in gene_batches]
        all_file_paths_lists = [f.result() for f in futures]

    gene_files = [path for sublist in all_file_paths_lists for path in sublist]

    # 8. Zip and Cleanup
    with zipfile.ZipFile(zip_output_path, "w", zipfile.ZIP_DEFLATED) as zf:
        zf.write(obs_path)
        zf.write("umap.parquet")
        for file_path in gene_files:
            zf.write(file_path)

    # Cleanup temp files
    os.remove(obs_path)
    os.remove("umap.parquet")
    for f_path in gene_files:
        os.remove(f_path)
    os.rmdir("genes_parquet")

    print(f"Success! Saved to {zip_output_path}")



##########---------------------------------------------------------------------
#ATAC
##########---------------------------------------------------------------------


def processing_atac_samples(atac_all_cells,
                            atlas_source_path = '/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
                            target_folder='/content/drive/MyDrive/pituitary_atlas/Unstructured/epitome_h5_files_atac_0911_final/',
                            target_folder_pdfs="/content/drive/MyDrive/pituitary_atlas/Unstructured/epitome_pdf_files_atac_0911_final/",
                            output_zip_name_path = "/content/drive/MyDrive/pituitary_atlas/Unstructured/epitome_h5_files_atac_0911_final.zip",
                            output_zip_name_pdfs_path = "/content/drive/MyDrive/pituitary_atlas/Unstructured/epitome_pdf_files_atac_0911_final.zip"

):

  pituitary_atlas = pd.read_excel(atlas_source_path)
  #make this folder if doesnt exist
  if not os.path.exists(target_folder):
      os.makedirs(target_folder)

  if not os.path.exists(target_folder_pdfs):
      os.makedirs(target_folder_pdfs)

  for geo in df["GEO"].values:

    print(geo)
    #try:
    adata = atac_all_cells[atac_all_cells.obs["sample"]==geo].copy()
    adata.obs["GEO"] = adata.obs["sample"]

    n_cells = adata.shape[0]
    print(adata.shape)
    #find this GEO value in pituitary_atlas, and add this to the col n_cells
    pituitary_atlas.loc[pituitary_atlas["GEO"] == geo, "n_cells"] = n_cells
    pituitary_atlas.loc[pituitary_atlas["GEO"] == geo, "most_recent_workflow"] = "v_0.01"

    adata.obs = adata.obs.merge(pituitary_atlas, on="GEO").copy()

    print(adata.shape)
    #write adata
    # Convert all columns in obs to strings if needed
    for col in adata.obs.columns:
        adata.obs[col] = adata.obs[col].astype(str)

    #change index to str
    adata.obsm["_scvi_extra_categorical_covs"].index = ["barcode_" + str(i) for i in adata.obs.index.astype(int)]
    adata.obs.index = ["barcode_" + str(i) for i in adata.obs.index.astype(int)]

    adata = adata.copy()
    #write adata
    adata.write(target_folder+geo+".h5ad")

    # save atac PDFs
    try:
      pdf_path = "/content/drive/MyDrive/pituitary_atlas/Unstructured/from_pc_0910/"+geo+"_summary.pdf"
      shutil.copy(pdf_path, target_folder_pdfs+geo+".pdf")
    except:
      print("No pdf found for "+geo)

    #except:
      print("Error reading "+geo)
      pituitary_atlas.loc[pituitary_atlas["GEO"] == geo, "n_cells"] = 0
      pituitary_atlas.loc[pituitary_atlas["GEO"] == geo, "most_recent_workflow"] = "none"
      continue

  #resave back to atlas
  pituitary_atlas.to_excel('/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx', index=False)


  
  # Zip the folder
  shutil.make_archive(output_zip_name_path.replace('.zip', ''), 'zip', target_folder)
  shutil.make_archive(output_zip_name_pdfs_path.replace('.zip', ''), 'zip', target_folder_pdfs)
  print(f"Zipped h5 files to {output_zip_name_path}")
  print(f"Zipped pdf files to {output_zip_name_pdfs_path}")


##############
#### Pseudobulking RNA
##############

def generate_pseudobulks(atlas_source_path = '/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
                         modality = ['sc','sn','multi_rna'],
                         min_cells_threshold = 300,
                         scvi_labels_to_keep = None,
                         filtering_based_on_barcodes = True,
                         ID_col = 'SRA_ID',
                         nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
                         remaining_path="/analysis/adata_new_assigned_0828.h5ad",
                         remaining_path_backup="/analysis/adata_assigned.h5ad",
                         sample_col='sample',
                          groups_col='new_cell_type',
                          min_cells=50,
                          pdata_output_path_suffix="/analysis/pdata_assigned_0828.h5ad",
                          include_new_samples = False

                         ):

  df = pd.read_excel(atlas_source_path)
  #modality
  df = df[df['Modality'].isin(modality)]
  #reset
  df.reset_index(drop=True, inplace=True)

  missing = []

  for i in range(len(df)-1,-1,-1):
    SRA_ID = df.iloc[i]['SRA_ID']
    GEO = df.iloc[i]['GEO']
    ID = df.iloc[i][ID_col]
    core_value = df[df[ID_col]==ID]["Core"].values[0]
    core_action = False
    if include_new_samples and core_value == 2:
      core_action = True

    print(ID)

    author = df.iloc[i]['Author']
    normal = df.iloc[i]['Normal']

    #if normal == 2 then skip
    if normal == 2:
      continue
    #if ["n_cells"]<300 skip
    n_cells = df.iloc[i]['n_cells']
    if n_cells < min_cells_threshold:
      continue
    
    #if f"{nac_path}{ID}{pdata_output_path_suffix}" exists
    if os.path.exists(f"{nac_path}{ID}{pdata_output_path_suffix}"):
      print(f"Skipping {ID} as pdata already exists")
      continue

     #read adata assigned
    try:
      print(f"Processing {ID}...")
      try:
        adata_assigned = sc.read(f"{nac_path}{ID}{remaining_path}")
      except:
        print("Falling back to adata_assigned")
        adata_assigned = sc.read(f"{nac_path}{ID}{remaining_path_backup}")
    except:
       print(f" Missing entry for {ID}")
       missing.append(ID)
       continue
    
    adata_assigned.obs["sample"] = ID
    cell_counts = adata_assigned.X.sum(axis=1)

    import matplotlib.pyplot as plt
    plt.hist(cell_counts,color='red')
    plt.show()

    if filtering_based_on_barcodes and core_action==False:

      if author == "Rebboah et al. (2025)":

        adata_assigned.obs["barcode"] = [GEO + ":" + adata_assigned.obs["bc"][i] + "_" + adata_assigned.obs["subpool"][i] + "-1"  for i in range(len(adata_assigned.obs))]
        shape1 = adata_assigned.shape
        adata_assigned = adata_assigned[adata_assigned.obs["barcode"].isin(scvi_labels_to_keep[scvi_labels_to_keep["SRA_ID"]==SRA_ID]["barcode"].values)]
        shape2 = adata_assigned.shape
        print(f"Removed {shape1[0]-shape2[0]} cells")


      else:
        adata_assigned.obs["barcode"] = [GEO+":" + adata_assigned.obs['0'][i] + "-1" for i in range(len(adata_assigned.obs))]
        shape1 = adata_assigned.shape
        adata_assigned = adata_assigned[adata_assigned.obs["barcode"].isin(scvi_labels_to_keep[scvi_labels_to_keep["SRA_ID"]==SRA_ID]["barcode"].values)]
        shape2 = adata_assigned.shape
        print(f"Removed {shape1[0]-shape2[0]} cells")

    #if sample_col not in obs add it 
    if sample_col not in adata_assigned.obs.columns:
      adata_assigned.obs[sample_col] = ID

    
    pdata_assigned = dc.get_pseudobulk(
        adata_assigned,
        sample_col=sample_col,
        groups_col=groups_col,
        mode='sum',
        min_cells=50,
        min_counts=0
    )

    pdata_assigned.obs.columns = pdata_assigned.obs.columns.astype(str)
    pdata_assigned.var.columns = pdata_assigned.var.columns.astype(str)

    for col in pdata_assigned.obs.columns:
      pdata_assigned.obs[col] = pdata_assigned.obs[col].astype(str)
    pdata_assigned.write(f"{nac_path}{ID}{pdata_output_path_suffix}")
    print(f"ALL DONE for {ID}")
  
    
  print("Missing assignments files for the following SRA_IDs:")
  print(missing)





def read_merge_save_pbulk(
      atlas_source_path = '/content/drive/MyDrive/pituitary_atlas/source_table/pituitary_atlas.xlsx',
      output_path="/content/drive/MyDrive/pituitary_atlas/Unstructured/pdatas0828.h5ad",
      Modality = ['sc','sn','multi_rna'],
      total_n_cells_threshold = 200,
      nac_path="/content/drive/MyDrive/pituitary_atlas/processed/nac/",
      remaining_path="/analysis/pdata_assigned_0828.h5ad",
      cell_type_col='new_cell_type',
      backup_cell_type_col='assignments'
      ):
  
  df = pd.read_excel(atlas_source_path)
  #in df keep only sc sn or multi_rna
  df = df[df['Modality'].isin(Modality)]
  #reset
  df.reset_index(drop=True, inplace=True)

  pdatas= []
  for i in range(len(df)):
    normal = df.iloc[i]['Normal']
    #if normal == 2 then skip
    if normal == 2:
      continue
    #if ["n_cells"]<300 skip
    n_cells = df.iloc[i]['n_cells']
    if n_cells < total_n_cells_threshold:
      continue
    SRA_ID = df.iloc[i]['SRA_ID']
    print(SRA_ID)

    try:
      pdata_assigned = sc.read(f"{nac_path}{SRA_ID}{remaining_path}")
      pdatas.append(pdata_assigned)
    except:
      print(f"Error reading {SRA_ID}")
      continue

  print(len(pdatas))

  pdatas_merged = anndata.concat(pdatas, join='outer')
  #change nans to 0 in pdatas
  pdatas_merged.X[np.isnan(pdatas_merged.X)] = 0

  #turn ["psbulk_n_cells"] to ints
  pdatas_merged.obs["psbulk_n_cells"] = pdatas_merged.obs["psbulk_n_cells"].astype(float)
  pdatas_merged.obs["psbulk_n_cells"].sum()

  # 1. Fill 'new_cell_type' where it equals the string "nan"
  #if there is backup_cell_type_col use that to fill in
  if backup_cell_type_col not in pdatas_merged.obs.columns:
    pdatas_merged.obs[backup_cell_type_col] = pdatas_merged.obs[cell_type_col]
  mask_cell_type = pdatas_merged.obs[cell_type_col].isna()
  pdatas_merged.obs.loc[mask_cell_type, cell_type_col] = \
      pdatas_merged.obs.loc[mask_cell_type, backup_cell_type_col]

  # 2. Fill 'sample' where it equals the string "nan"
  mask_sample = pdatas_merged.obs["sample"].isna()
  pdatas_merged.obs.loc[mask_sample, "sample"] = \
      pdatas_merged.obs.loc[mask_sample, "SRA_ID"]

  #add metadata from df
  #only keep sample and new_cell_type
  pdatas_merged.obs =pdatas_merged.obs[['sample',cell_type_col,"psbulk_n_cells"]]
  #rename SRA_ID to sample
  pdatas_merged.obs = pd.merge(pdatas_merged.obs, df, left_on='sample', right_on='SRA_ID')

  #from pdatas.obs remove columns that contain "Unnamed" or _index
  pdatas_merged.obs = pdatas_merged.obs.loc[:, ~pdatas_merged.obs.columns.str.contains('Unnamed')]
  pdatas_merged.obs = pdatas_merged.obs.loc[:, ~pdatas_merged.obs.columns.str.contains('_index')]

  pdatas_merged.obs.columns = pdatas_merged.obs.columns.astype(str)
  pdatas_merged.var.columns = pdatas_merged.var.columns.astype(str)
  pdatas_merged.var.index = pdatas_merged.var.index.astype(str)
  pdatas_merged.obs.index= pdatas_merged.obs.index.astype(str)

  # Convert all columns in obs to strings if needed
  for col in pdatas_merged.obs.columns:
      if pdatas_merged.obs[col].apply(type).nunique() > 1:
          pdatas_merged.obs[col] = pdatas_merged.obs[col].astype(str)

  # Now, try saving the AnnData object
  pdatas_merged.write(f"{output_path}")

















import xgboost as xgb
import joblib
from pathlib import Path
import scipy.sparse as sp
import numpy as np
import scanpy as sc
import pandas as pd
from collections import Counter

from importlib import resources
from pathlib import Path

def get_cell_typing_path(
    version: str,
    species: str,
    modality: str,
) -> tuple[Path, Path]:
    """
    Resolve packaged doublet model paths inside epitome_tools.
    
    Returns:
        model_path: Path to the XGBoost JSON model
        label_encoder_path: Path to the label encoder (joblib)
    """
    if modality == "rna":
        task = "cell_typing_rna"
        model_file = "rna_model.json"
        label_encoder_file = "rna_label_encoder.pkl"
    elif modality == "atac":
        task = "cell_typing_atac"
        model_file = "atac_model.json"
        label_encoder_file = "atac_label_encoder.pkl"
    else:
        raise ValueError(f"Unsupported modality: {modality}")

    base_path = resources.files("epitome_tools") / "models" / version / species / task

    model_path = base_path / model_file
    label_encoder_path = base_path / label_encoder_file

    # Optional sanity check
    for path in (model_path, label_encoder_path):
        if not path.exists():
            raise FileNotFoundError(f"Packaged file not found: {path}")

    return model_path, label_encoder_path


def load_celltype_model(version="v_0.02",species="mouse", modality="rna"):
    """
    Load the XGBoost model from the specified path.
    """
    model_path, label_encoder_path = get_cell_typing_path(version, species, modality)
    model = xgb.XGBClassifier()
    model.load_model(model_path)

    # Access the booster and retrieve the feature names
    booster = model.get_booster()

    # Get the feature names (these should match the features used for training)
    feature_names = model.feature_names_in_

    label_encoder = joblib.load(label_encoder_path)

    return model, label_encoder, feature_names



def prepare_matrix_celltype(adata, feature_names,active_assay="sc", nan_or_zero='nan'):
    """
    Prepare the AnnData object for prediction by selecting the relevant features.
    """

    adata_to_handle = adata.copy()

    # Define all potential assay features based on feature_names_sorted1
    assay_features = [f for f in feature_names if f.startswith('assay_')]

    #number of cells
    n_obs = adata_to_handle.n_obs

    #initialise assay data with zeros
    assay_data1 = np.zeros((n_obs, len(assay_features)), dtype=np.float32)

    # Create a mapping from assay feature name to its column index
    assay_feature_indices1 = {name: i for i, name in enumerate(assay_features)}

    #active assay must be one of sc, sn, multi_rna, if neither, set it as sc
    if active_assay not in ['sc', 'sn', 'multi_rna']:
        print(f"Warning: Active assay '{active_assay}' not recognized. Defaulting to 'sc'.")
        active_assay = 'sc'
    #add prefix
    active_assay = f"assay_{active_assay}_onehot"

    if active_assay in assay_feature_indices1:
        assay_data1[:, assay_feature_indices1[active_assay]] = 1.0

    # Convert assay data to sparse if original data is sparse
    assay_data_matrix1 = sp.csr_matrix(assay_data1) if sp.issparse(adata_to_handle.X) else assay_data1

    # --- 2. Combine Gene Expression and Assay Features for Model 1 ---
    # We need to combine these temporarily to easily subset later

    # Ensure original data is CSR for efficient column slicing if sparse
    if sp.issparse(adata_to_handle.X) and not isinstance(adata_to_handle.X, sp.csr_matrix):
        adata_X = adata_to_handle.X.tocsr().copy()
        print("Converted adata_orig.X to CSR format.")
    else:
        adata_X = adata_to_handle.X.copy()

    # Combine the matrices horizontally
    combined_X1 = sp.hstack([adata_X, assay_data_matrix1], format='csr') if sp.issparse(adata_X) else np.hstack([adata_X, assay_data1])

    # Create combined feature names list
    combined_feature_names1 = adata_to_handle.var_names.tolist() + assay_features

    # Create a mapping from the combined feature names to their column index
    combined_feature_indices1 = {name: i for i, name in enumerate(combined_feature_names1)}
    print(f"Combined matrix shape for model 1: {combined_X1.shape}")

    # Initialize the final matrix with NaNs (XGBoost can handle NaNs)
    if nan_or_zero == 'nan':
        X_final1 = np.full((n_obs, len(feature_names)), np.nan, dtype=np.float32)
    elif nan_or_zero == 'zero':
        X_final1 = np.zeros((n_obs, len(feature_names)), dtype=np.float32)

    # Create a mapping for the target feature order
    target_feature_indices1 = {name: i for i, name in enumerate(feature_names)}

    # Find which features required by the model are present in our combined data
    available_features1 = [f for f in feature_names if f in combined_feature_indices1]
    missing_features1 = [f for f in feature_names if f not in combined_feature_indices1]

    if missing_features1:
        print(f"Warning: {len(missing_features1)} features required by model 1 are missing from the data: {missing_features1[:5]}...") # Print first 5

    print(f"Found {len(available_features1)} available features out of {len(feature_names)} required for model 1.")

    # Get the column indices in the *combined* data for the available features - indices of where it is in the combined matrix
    source_indices1 = [combined_feature_indices1[f] for f in available_features1]

    # Get the column indices in the *final* matrix for these available features - indices of where the model expects it
    target_indices1 = [target_feature_indices1[f] for f in available_features1]

    # Fill the final matrix with data from the available features
    # Ensure data is dense for assignment; handle potential memory issues for large datasets
    if sp.issparse(combined_X1):
        # Slice sparse matrix efficiently and convert to dense for assignment
        X_final1[:, target_indices1] = combined_X1[:, source_indices1].toarray()
        print("Filled final matrix for model 1 from sparse data.")
    else:
        X_final1[:, target_indices1] = combined_X1[:, source_indices1]
        print("Filled final matrix for model 1 from dense data.")
    
    
    return X_final1


def perform_celltype_prediction(matrix, model, label_encoder, return_probas=True):
    """
    Perform cell type prediction using the provided model and label encoder.
    """
    probas = model.predict_proba(matrix)
    predicted_labels = model.predict(matrix)
    predicted_cell_types = label_encoder.inverse_transform(predicted_labels)

    if return_probas:
        return predicted_cell_types, probas
    else:
        return predicted_cell_types
    

def smoothing_cell_types(adata):
    """
    Checks for PCA embedding, generates if missing, finds 10 closest neighbors,
    and reassigns cell identity based on neighbor majority.
    """
    if 'X_pca' not in adata.obsm:
        sc.tl.pca(adata)

    sc.pp.neighbors(adata, n_neighbors=10, use_rep='X_pca')

    new_cell_types = adata.obs['predicted_cell_type'].copy()

    for i in range(adata.n_obs):
        # Access neighbor indices from the connectivities matrix in adata.obsp
        neighbor_indices = adata.obsp['connectivities'].indices[
            adata.obsp['connectivities'].indptr[i]:
            adata.obsp['connectivities'].indptr[i+1]
        ]
        neighbor_cell_types = adata.obs['predicted_cell_type'].iloc[neighbor_indices]
        type_counts = Counter(neighbor_cell_types)

        for cell_type, count in type_counts.items():
            if count > 5:
                new_cell_types.iloc[i] = cell_type
                break

    adata.obs['cell_type_final'] = new_cell_types
    return adata


import numpy as np
import scipy.sparse as sp
import joblib
import xgboost as xgb
from importlib import resources
from pathlib import Path

from importlib import resources
from pathlib import Path

def get_doublet_path(
    version: str,
    species: str,
    modality: str,
) -> tuple[Path, Path, Path]:
    """
    Resolve packaged doublet model paths inside epitome_tools.
    
    Returns:
        model_path: Path to the XGBoost JSON model
        label_encoder_path: Path to the label encoder (joblib)
        threshold_path: Path to the threshold (joblib)
    """
    if modality == "rna":
        task = "doublet_rna"
        model_file = "rna_model_binary.json"
        label_encoder_file = "rna_label_encoder_binary.pkl"
        threshold_file = "final_threshold.pkl"
    elif modality == "atac":
        task = "doublet_atac"
        model_file = "atac_model_binary.json"
        label_encoder_file = "atac_label_encoder_binary.pkl"
        threshold_file = "final_threshold.pkl"
    else:
        raise ValueError(f"Unsupported modality: {modality}")

    base_path = resources.files("epitome_tools") / "models" / version / species / task

    model_path = base_path / model_file
    label_encoder_path = base_path / label_encoder_file
    threshold_path = base_path / threshold_file

    # Optional sanity check
    for path in (model_path, label_encoder_path, threshold_path):
        if not path.exists():
            raise FileNotFoundError(f"Packaged file not found: {path}")

    return model_path, label_encoder_path, threshold_path


def load_doublet_model(version="v_0.02",species="mouse", modality="rna"):
    """
    Load the XGBoost model for doublet prediction from the specified path.
    """

    model_path, label_encoder_path, threshold_path = get_doublet_path(version, species, modality)
    model = xgb.XGBClassifier()
    model.load_model(model_path)

    # Access the booster and retrieve the feature names
    booster = model.get_booster()

    # Get the feature names (these should match the features used for training)
    feature_names = model.feature_names_in_

    label_encoder = joblib.load(label_encoder_path)
    
    threshold = joblib.load(threshold_path)

    return model, label_encoder, threshold, feature_names


def prepare_matrix_doublet(adata, feature_names,  active_assay="sc", nan_or_zero='nan'):

    adata_to_handle = adata.copy()

    assay_features1 = [f for f in feature_names if f.startswith('pct_counts_kept')]
    assay_features2 = [f for f in feature_names if f.endswith('one_hot')]

    assay_features = assay_features1 + assay_features2

    #number of cells
    n_obs = adata_to_handle.n_obs

    #initialise assay data with zeros
    assay_data = np.zeros((n_obs, len(assay_features)), dtype=np.float32)

    # Create a mapping from assay feature name to its column index
    assay_feature_indices = {name: i for i, name in enumerate(assay_features)}

    #active assay must be one of sc, sn, multi_rna, if neither, set it as sc
    if active_assay not in ['sc', 'sn', 'multi_rna']:
        print(f"Warning: Active assay '{active_assay}' not recognized. Defaulting to 'sc'.")
        active_assay = 'sc'
        #add as prefix
    active_assay = f"{active_assay}_one_hot"


    if active_assay in assay_feature_indices:
        assay_data[:, assay_feature_indices[active_assay]] = 1.0

    #add adata_to_handle.obs['pct_counts_kept'] if exists
    if 'pct_counts_kept' in adata_to_handle.obs.columns and 'pct_counts_kept' in assay_feature_indices:
        print("Adding 'pct_counts_kept' to assay data.")
        #print first 5 values
        print(adata_to_handle.obs['pct_counts_kept'].values[:5])
        assay_data[:, assay_feature_indices['pct_counts_kept']] = adata_to_handle.obs['pct_counts_kept'].values.astype(np.float32)

    # Convert assay data to sparse if original data is sparse
    assay_data_matrix = sp.csr_matrix(assay_data) if sp.issparse(adata_to_handle.X) else assay_data

    # --- 2. Combine Gene Expression and Assay Features for Model 1 ---
    # We need to combine these temporarily to easily subset later

    # Ensure original data is CSR for efficient column slicing if sparse
    if sp.issparse(adata_to_handle.X) and not isinstance(adata_to_handle.X, sp.csr_matrix):
        adata_X = adata_to_handle.X.tocsr().copy()
        print("Converted adata_orig.X to CSR format.")
    else:
        adata_X = adata_to_handle.X.copy()

    # Combine the matrices horizontally
    combined_X1 = sp.hstack([adata_X, assay_data_matrix], format='csr') if sp.issparse(adata_X) else np.hstack([adata_X, assay_data])

    # Create combined feature names list
    combined_feature_names = adata_to_handle.var_names.tolist() + assay_features

    # Create a mapping from the combined feature names to their column index
    combined_feature_indices = {name: i for i, name in enumerate(combined_feature_names)}
    print(f"Combined matrix shape for model 1: {combined_X1.shape}")

    if nan_or_zero == 'nan':
        X_final = np.full((n_obs, len(feature_names)), np.nan, dtype=np.float32)
    elif nan_or_zero == 'zero':
        X_final = np.zeros((n_obs, len(feature_names)), dtype=np.float32)
        
    target_feature_indices = {name: i for i, name in enumerate(feature_names)}

    # Reuse combined_feature_indices1, available_features1, source_indices1 from Model 1
    available_features = [f for f in feature_names if f in combined_feature_indices] 
    missing_features = [f for f in feature_names if f not in combined_feature_indices]

    if missing_features:
        print(f"Warning: {len(missing_features)} features required by model 2 are missing: {missing_features[:5]}...")

    print(f"Found {len(available_features)} available features out of {len(feature_names)} required for model 2.")
    source_indices = [combined_feature_indices[f] for f in available_features] # Use combined_feature_indices1

    target_indices = [target_feature_indices[f] for f in available_features]

    if sp.issparse(combined_X1):
        # Slice sparse matrix efficiently and convert to dense for assignment
        X_final[:, target_indices] = combined_X1[:, source_indices].toarray()
        print("Filled final matrix for model 1 from sparse data.")
    else:
        X_final[:, target_indices] = combined_X1[:, source_indices]
        print("Filled final matrix for model 1 from dense data.")
    return X_final

def perform_doublet_prediction(matrix, model, label_encoder, threshold):

    n_obs = matrix.shape[0]
    predicted_labels = model.predict(matrix)
    predicted_doublet_labels = label_encoder.inverse_transform(predicted_labels)
    probas = model.predict_proba(matrix) # Get probabilities
    print("Doublet prediction with model complete.")
    is_doublet = np.full(n_obs, True, dtype=bool) # Default: not a doublet
    doublet_score = np.zeros(n_obs)

    if probas.shape[1] > 1:
        doublet_probabilities = probas[:, 1]
    elif probas.shape[1] == 1:
        doublet_probabilities = probas[:, 0]
    else:
        doublet_probabilities = np.zeros(n_obs)
        print("Warning: Model 2 probas has shape < 1")

    for i in range(n_obs):
        doublet_score[i] = 1 - doublet_probabilities[i]
        if doublet_probabilities[i] > (1-threshold):
            is_doublet[i] = False


    return predicted_doublet_labels, is_doublet, doublet_score


import subprocess
import os
import pandas as pd
import subprocess
import json
from scipy.io import mmread, mmwrite
import numpy as np
from scipy.sparse import csc_matrix
from scipy.stats import median_abs_deviation as mad
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import tempfile
import scipy.sparse as sp
import scanpy as sc
import time

def sra_downloader(raw_path,df):
    """
    Downloads SRA datasets from NCBI using wget and ffq.
    Args:
        raw_path (str): Path to the directory where datasets will be downloaded.
        df (pd.DataFrame): DataFrame containing SRA IDs and experiment types.
    """

    datasets_already_downloaded = os.listdir(raw_path)

    #iterating through the table and getting datasets
    for i in range(len(df)-1,-1,-1):
        SRA_ID = df.iloc[i]['SRA_ID']
        experiment_type = df.iloc[i]['Modality']

        if SRA_ID in datasets_already_downloaded:
            print("SRA_ID already downloaded")
            continue

        if (not os.path.exists(raw_path+ SRA_ID)) and ((experiment_type == "sc") or (experiment_type == "sn")):
            os.mkdir(raw_path + SRA_ID)

            print("Folder created")

        target_folder = raw_path + SRA_ID
        download_command = f"wget -P {target_folder} /path/to/target/folder $(ffq --ncbi {SRA_ID} | jq -r '.[] | .url' | tr '\n' ' ')"
        #run command
        print(download_command)
        subprocess.run(download_command, shell=True)


def fastq_downloader(raw_path, df):
    """
    Downloads fastq files from SRA datasets.
    Args:
        raw_path (str): Path to the directory where datasets are stored.
        df (pd.DataFrame): DataFrame containing SRA IDs and experiment types.
    """

    for i in range(len(df)-1,-1,-1):
        SRA_ID = df.iloc[i]['SRA_ID']
        experiment_type = df.iloc[i]['Modality']
        outdir = raw_path + SRA_ID
        files = os.listdir(outdir)
        for file in files:
            print(file)
            if ("lite" in file) or ("sra" in file):
                #check if there are fastq files with starting with file
                if any(x.startswith(file) and x.endswith('.fastq.gz') for x in files):
                    print("fastq files found")
                    #or it has bam string in it
                elif "bam" in file:
                    print("bam file found")
                else:
                    file = raw_path + SRA_ID + '/' + file
                    fastqdump_command= f"fastq-dump --outdir {outdir} --gzip --split-files {file}"
                    subprocess.run(fastqdump_command, shell=True)


def run_kb_count_nac(nac_path,raw_path, df, species="mouse"):
    """ Runs kb count for nascent RNA datasets.
    Args:
        nac_path (str): Path to the directory where nascent RNA datasets will be stored.
        raw_path (str): Path to the directory where raw datasets are stored.
        df (pd.DataFrame): DataFrame containing SRA IDs, experiment types, and technologies.
    """

    kbref_command_nac = f"kb ref --workflow=nac  -d {species} -i index_nac.idx -g t2g_nac.txt -g t2g_nac.txt -c1 cdna.txt -c2 nascent.txt -f1 cdna.fasta -f2 nascent.fasta --overwrite"
    kbref_command_std = f"kb ref -d {species} -i index_std.idx -g t2g_std.txt --overwrite"

    #if /index_nac.idx doesnt exist, run kb ref
    if not os.path.exists("index_nac.idx"):
        print("Running kb ref for nac")
        subprocess.run(kbref_command_nac, shell=True)

    if not os.path.exists("index_std.idx"):
        print("Running kb ref for std")
        subprocess.run(kbref_command_std, shell=True)

    datasets_already_downloaded = os.listdir(nac_path)
    
    #iterating through the table and getting datasets
    for i in range(len(df)):
        SRA_ID = df.iloc[i]['SRA_ID']
        experiment_type = df.iloc[i]['Modality']
        tech = df.iloc[i]['10X version']
        author = df.iloc[i]['Author']

        #print all of these
        print("""
        SRA_ID: {}
        experiment_type: {}
        tech: {}
        """.format(SRA_ID,experiment_type,tech))

        #check if SRA_ID is already downloaded
        if SRA_ID in datasets_already_downloaded:
            print("SRA_ID already aligned for nac")
            continue
        
        all_fastqs = find_fastqs(raw_path, SRA_ID)
        #if there arent at least 2 fastqs then continue
        if len(all_fastqs) < 2:
            print("Not enough fastqs")
            continue

        if not os.path.exists(nac_path + SRA_ID):
            os.mkdir(nac_path + SRA_ID)

        if experiment_type == "sn":
            print("Running kb count")
            output_dir = f"{nac_path}{SRA_ID}/"
            kbcount_nac_command1 = f"kb count -i index_nac.idx -g t2g_nac.txt -c1 cdna.txt -c2 nascent.txt  -x {tech} --workflow=nac --sum=total --overwrite --h5ad -o {output_dir} {' '.join(all_fastqs)}"
            print(kbcount_nac_command1)
            subprocess.run(kbcount_nac_command1, shell=True)

        elif (experiment_type == "sc"):
            print("Running kb count")
            output_dir = f"{nac_path}{SRA_ID}/"
            kbcount_nac_command1 = f"kb count -i index_nac.idx -g t2g_nac.txt -c1 cdna.txt -c2 nascent.txt -x {tech} --workflow=nac --sum=total --overwrite --h5ad -o {output_dir} {' '.join(all_fastqs)}"
            print(kbcount_nac_command1)
            subprocess.run(kbcount_nac_command1, shell=True)
            
            #change nac to tcc
            output_dir_tcc = nac_path.replace("nac", "tcc")
            #make sure it exists
            if not os.path.exists(output_dir_tcc):
                os.mkdir(output_dir_tcc)
            kbcount_tcc_command = f"kb count -i index_std.idx -g t2g_std.txt --overwrite -x {tech} --tcc --h5ad -o {output_dir_tcc} {' '.join(all_fastqs)}"
            print(kbcount_tcc_command)
            subprocess.run(kbcount_tcc_command, shell=True)

        elif (experiment_type == "multi_rna"):
            print("Running kb count")
            output_dir = f"{nac_path}{SRA_ID}/"
            #from all_fastqs remove ones that contain "atac"
            all_fastqs = [x for x in all_fastqs if "atac" not in x]
            print(all_fastqs)
            kbcount_nac_command1 = f"kb count -i index_nac.idx -g t2g_nac.txt -c1 cdna.txt -c2 nascent.txt -x {tech} -w /10xMultiome/gex_737K-arc-v1.txt --workflow=nac --sum=total --overwrite --h5ad -o {output_dir} {' '.join(all_fastqs)}"
            print(kbcount_nac_command1)
            subprocess.run(kbcount_nac_command1, shell=True)

def find_fastqs(raw_path, SRA_ID):
    """ Finds fastq files in the specified directory for a given SRA ID.
    Args:
        raw_path (str): Path to the directory where datasets are stored.
        SRA_ID (str): SRA ID for which fastq files are to be found.
    Returns:
        list: List of fastq file paths.
    """

    directory = raw_path + SRA_ID
    files = os.listdir(directory)
    all_fastqs = []
    for file in files:
        if file.endswith('.fastq.gz'):
            all_fastqs.append(directory + '/' + file)

    lite_name = True if any(x for x in all_fastqs if 'lite' in x) else False
    if lite_name:
        id_tags = [x.split('lite')[0] for x in all_fastqs]

    SRR_name = True if any(x for x in all_fastqs if 'SRR' in x) else False
    if SRR_name:
        id_tags = [x.split('.1_')[0] for x in all_fastqs]

        if id_tags[0].count('/') > 1:
            id_tags = [x.split('_1')[0] for x in all_fastqs]

    #if neither
    elif (not lite_name) and (not SRR_name):
        id_tags = [x.split('_R1_')[0] for x in all_fastqs if '_R1_' in x]

    print(id_tags)
    all_fastqs_final = []
    unique_tags= list(set(id_tags))
    for tag in unique_tags:
        fastqs_with_tag = [x for x in all_fastqs if tag in x]
        if len(fastqs_with_tag) > 1:
            #sort by size and take the two largest
            fastqs_with_tag.sort(key=lambda x: os.path.getsize(x), reverse=True)
            all_fastqs_final.append(fastqs_with_tag[1])
            all_fastqs_final.append(fastqs_with_tag[0])

    all_fastqs = all_fastqs_final
    return all_fastqs


def process_matrix_files(matrix_file, barcodes_file,matrix_file_new,barcodes_file_new,num=10):
    # Read matrix data
    # Load matrix data
    matrix_data = mmread(matrix_file).tocsc()  # Load matrix as compressed sparse column format

    # Read barcodes
    barcodes = pd.read_csv(barcodes_file, header=None)
    #squeeze
    barcodes = barcodes.squeeze()

    # Compute the sum of counts for each cell (column)
    cell_sums = matrix_data.sum(axis=1).A1

    # Filter cells where the total counts are greater than 10,000
    filtered_cell_indices = cell_sums > num

    filtered_matrix_data = matrix_data[filtered_cell_indices, :]
    filtered_barcodes = barcodes[filtered_cell_indices]

    # Save the filtered matrix
    mmwrite(matrix_file_new, csc_matrix(filtered_matrix_data))

    # Save the filtered barcodes
    filtered_barcodes.to_csv(barcodes_file_new, header=False, index=False)

    # Load Matrix Market file as DataFrame
    matrix_data_df = pd.read_csv(matrix_file_new, sep=' ', header=2)

    # Sort by first then second column
    first_col = matrix_data_df.columns[0]
    second_col = matrix_data_df.columns[1]
    matrix_data_sorted = matrix_data_df.sort_values(by=[first_col, second_col])

    # Write back to .mtx file with the same %header as before
    # Write header lines
    with open(matrix_file_new, 'r') as f:
        header_lines = [f.readline() for _ in range(2)]  # Read the first three lines

    # Write sorted data to the matrix file
    with open(matrix_file_new, 'w') as f:
        f.writelines(header_lines)
        matrix_data_sorted.to_csv(f, sep=' ', index=False, header=True, float_format='%.6g')

    print(f"Sorted Matrix Market file saved as: {matrix_file_new}")


def filter_top_cells_mtx(matrix_file, barcodes_file, output_matrix_file, output_barcodes_file, top_n=50000):
    """
    Filter MTX matrix and corresponding barcode file to keep only the top N cells by total counts.
    Filters rows of the matrix (cells) while keeping all columns (genes).

    Parameters:
    -----------
    matrix_file : str
        Path to input MTX file
    barcodes_file : str
        Path to input barcodes file
    output_matrix_file : str
        Path to save filtered MTX file
    output_barcodes_file : str
        Path to save filtered barcodes file
    top_n : int, optional
        Number of top cells to keep (default: 50000)

    Returns:
    --------
    tuple
        (filtered_matrix, filtered_barcodes)
    """
    # Load the matrix in CSR format for efficient row operations
    matrix = mmread(matrix_file).tocsr()

    # Read barcodes
    barcodes = pd.read_csv(barcodes_file, header=None).squeeze()

    # Calculate row sums (total counts per cell)
    row_sums = np.array(matrix.sum(axis=1)).flatten()

    # Get indices of top N cells by total counts
    top_cell_indices = np.argsort(row_sums)[-top_n:]
    top_cell_indices.sort()  # Sort indices to maintain original order

    # Filter matrix and barcodes
    filtered_matrix = matrix[top_cell_indices, :]
    filtered_barcodes = barcodes.iloc[top_cell_indices]

    # Save filtered matrix
    mmwrite(output_matrix_file, filtered_matrix)

    # Save filtered barcodes
    filtered_barcodes.to_csv(output_barcodes_file, header=False, index=False)

    # Sort the matrix file if needed
    matrix_data_df = pd.read_csv(output_matrix_file, sep=' ', header=2)

    # Sort by first then second column
    first_col = matrix_data_df.columns[0]
    second_col = matrix_data_df.columns[1]
    matrix_data_sorted = matrix_data_df.sort_values(by=[first_col, second_col])

    # Write header lines and sorted data
    with open(output_matrix_file, 'r') as f:
        header_lines = [f.readline() for _ in range(2)]

    with open(output_matrix_file, 'w') as f:
        f.writelines(header_lines)
        matrix_data_sorted.to_csv(f, sep=' ', index=False, header=True, float_format='%.6g')

    return filtered_matrix, filtered_barcodes

def filtering_cells(df,raw_path="/content/drive/MyDrive/pituitary_atlas",root_path = "/content/",atlas_source_file = "pituitary_atlas.xlsx",force=False):
    max_retries = 3
    for i in range(len(df)-1,-1,-1):
        SRA_ID = df.iloc[i]['SRA_ID']
        if os.path.exists(f"{raw_path}/processed/nac/{SRA_ID}/analysis"):
            print(SRA_ID)
            try:
                #see if {raw_path}/processed/nac/{SRA_ID}/counts_unfiltered/ exists or continue
                if not os.path.exists(f"{raw_path}/processed/nac/{SRA_ID}/counts_unfiltered/"):
                    print(f"{raw_path}/processed/nac/{SRA_ID}/counts_unfiltered/ does not exist, skipping.")
                    continue
                barcodes= f"{raw_path}/processed/nac/{SRA_ID}/counts_unfiltered/cells_x_genes.barcodes.txt"
                matrix_file = f"{raw_path}/processed/nac/{SRA_ID}/counts_unfiltered/cells_x_genes.total.mtx"
                genes = f"{raw_path}/processed/nac/{SRA_ID}/counts_unfiltered/cells_x_genes.genes.names.txt"
                barcodes_out = f"{raw_path}/processed/nac/{SRA_ID}/analysis/out_42_cells_x_genes.barcodes.txt"
                output_mat = f"{raw_path}/processed/nac/{SRA_ID}/analysis/out_42_cells_x_genes.total.mtx"

                #if output_mat exists, continue
                if not force:
                    if os.path.exists(output_mat):
                        #set filtering_junk to 1
                        atlas =  pd.read_excel(f'{raw_path}/source_table/{atlas_source_file}')
                        atlas.loc[atlas['SRA_ID'] == SRA_ID, 'filtering_junk'] = 1
                        #save back
                        atlas.to_excel(f'{raw_path}/source_table/{atlas_source_file}', index=False)
                        print(f"{output_mat} already exists, skipping filtering.")
                        continue
                i =0

                while (not os.path.exists(barcodes_out) or not os.path.exists(output_mat)) and i < max_retries:
                    filter_top_cells_mtx(matrix_file, barcodes,output_mat,barcodes_out,40000)
                    #sleep 30 sec to allow for files to sync with google drive
                    time.sleep((i+1)*30)
                    i += 1
                
                
                #mx filter
                barcodes_out2 = f"{raw_path}/processed/nac/{SRA_ID}/analysis/out_43_cells_x_genes.barcodes.txt"
                output_mat2 = f"{raw_path}/processed/nac/{SRA_ID}/analysis/out_43_cells_x_genes.total.mtx"

                mxfilter_command = f"mx filter -c 2 -bi {barcodes_out} -bo {barcodes_out2} -o {output_mat2} {output_mat}"
                #run command
                for attempt in range(1, max_retries + 1):
                    result = subprocess.run(mxfilter_command, shell=True)

                    if result.returncode == 0:
                        print(f"Command succeeded on attempt {attempt}")
                        break
                    else:
                        print(f"Attempt {attempt} failed with return code {result.returncode}")
                else:
                    print("All attempts failed.")
                
                #inspect
                inspect_command = f"mx inspect -a all -o {root_path}inspect_check1.json -gi {genes} -bi {barcodes_out2} {output_mat2}"
                time.sleep(30) #sleep 30 sec to allow for files to sync with google drive
                subprocess.run(inspect_command, shell=True)
                time.sleep(30) #sleep 30 sec to allow for files to sync with google drive
                
                inspect=json.load(open(f'{root_path}inspect_check1.json'))
                ncells= inspect["ncells"]

                barcodes_out3 = f"{raw_path}/processed/nac/{SRA_ID}/analysis/out_44_cells_x_genes.barcodes.txt"
                output_mat3 = f"{raw_path}/processed/nac/{SRA_ID}/analysis/out_44_cells_x_genes.total.mtx"

                if ncells < 3000:
                    #sleep 30 sec to allow for files to sync with google drive
                    time.sleep(30) #sleep 30 sec to allow for files to sync with google drive
                    process_matrix_files(matrix_file, barcodes,output_mat3,barcodes_out3,num=1500)
                    time.sleep(30) #sleep 30 sec to allow for files to sync with google drive
                else:
                    i = 0
                    while not os.path.exists(barcodes_out3) or not os.path.exists(output_mat3) or i < max_retries:
                        #filter top cells
                        time.sleep(30) #sleep 30 sec to allow for files to sync with google drive
                        filter_top_cells_mtx(matrix_file, barcodes,output_mat3,barcodes_out3,int(int(ncells)+min(0.3*int(ncells),2000)))
                        time.sleep(30) #sleep 30 sec to allow for files to sync with google drive
                        i += 1

                barcodes_out4 = f"{raw_path}/processed/nac/{SRA_ID}/analysis/out_45_cells_x_genes.barcodes.txt"

                output_mat4 = f"{raw_path}/processed/nac/{SRA_ID}/analysis/out_45_cells_x_genes.total.mtx"

                genes_out = f"{raw_path}/processed/nac/{SRA_ID}/analysis/out_45_cells_x_genes.genes.names.txt"

                #if these all exist, print that we have successfully produced the output
                if os.path.exists(output_mat4) and os.path.exists(barcodes_out4) and os.path.exists(genes_out):
                    print(f"Successfully produced output for {SRA_ID}")

                if not force:
                    #if output_mat4  exist continue
                    if os.path.exists(output_mat4):
                        continue

                mxclean_command = f"mx clean -gi {genes} -go {genes_out} -bi {barcodes_out3} -bo {barcodes_out4} -o {output_mat4} {output_mat3}"
                for attempt in range(1, max_retries + 1):
                    time.sleep(30) #sleep 30 sec to allow for files to sync with google drive
                    result = subprocess.run(mxclean_command, shell=True)
                    time.sleep(30) #sleep 30 sec to allow for files to sync with google drive
                    if result.returncode == 0:
                        print(f"Command succeeded on attempt {attempt}")
                        break
                    else:
                        print(f"Attempt {attempt} failed with return code {result.returncode}")
                else:
                    print("All attempts failed.")

                #inspect
                
                inspect_command = f"mx inspect -a all -o {root_path}inspect_check1.json -gi {genes_out} -bi {barcodes_out4} {output_mat4}"
                time.sleep(30)
                subprocess.run(inspect_command, shell=True)
                time.sleep(30)
                #read json
                
                inspect=json.load(open(f'{root_path}inspect_check1.json'))

                #print ncells
                print("We have this many cells: ", inspect["ncells"])
                print("We have this many genes: ", inspect["ngenes"])

                #set filtering_junk to 1
                atlas =  pd.read_excel(f'{raw_path}/source_table/{atlas_source_file}')
                atlas.loc[atlas['SRA_ID'] == SRA_ID, 'filtering_junk'] = 1
                atlas.to_excel(f'{raw_path}/source_table/{atlas_source_file}', index=False)

            except Exception as e:
                #print explicit error
                print(f"Error for {SRA_ID}: {e}")


def qc1(adata,species,min_genes=800, min_counts=1000):
    #you could also use a whitelist of barcodes from the filtered barcodes for each sample - not even sure what I meant by this
    if species == "mouse":
      mt_tag="mt-"
      ribo_tag=("Rps","Rpl")
    elif species == "human":
      mt_tag="MT-"
      ribo_tag=("RPS","RPL")
    sc.pp.filter_cells(adata, min_genes = min_genes)
    sc.pp.filter_cells(adata, min_counts = min_counts)
    adata.var["mt"] = adata.var_names.str.startswith(mt_tag)
    adata.var["ribo"] = adata.var_names.str.startswith(ribo_tag)
    sc.pp.calculate_qc_metrics(adata, qc_vars=["mt", "ribo"], inplace=True, percent_top=[20], log1p=True)

    return adata

def mad_outlier(adata, metric, nmads, upper_only = False,value = False):
    if value and upper_only:
      upper_val = np.median(adata.obs[metric]) + nmads * mad(adata.obs[metric])
      return [upper_val,0]
    elif value and not upper_only:
      upper_val = np.median(adata.obs[metric]) + nmads * mad(adata.obs[metric])
      lower_val = np.median(adata.obs[metric]) - nmads * mad(adata.obs[metric])
      return [upper_val,lower_val]
    else:
      M = adata.obs[metric]

      if not upper_only:
          return (M < np.median(M) - nmads * mad(M)) | (M > np.median(M) + nmads * mad(M))

    return (M > np.median(M) + nmads * mad(M))

def violinplot(ax, adata, col, vals):
    values = adata.obs[col].values
    sns.violinplot(y=values, ax=ax)
    for v in vals:
        ax.axhline(v, color='red')
    ax.set_title(col)
    ax.set_ylabel('')

def qc2(adata, species, experiment="sc"):
    if experiment == "sc":
        if species == "mouse":
            malat_tag = "Malat1"
        elif species == "human":
            malat_tag = "MALAT1"

        adata.var["malat"] = adata.var_names.str.startswith(malat_tag)

        sc.pp.calculate_qc_metrics(adata, qc_vars=["malat"], inplace=True, percent_top=[20], log1p=True)

        mt_val = mad_outlier(adata, 'pct_counts_mt', 5, upper_only=True, value=True)
        ribo_val = mad_outlier(adata, 'pct_counts_ribo', 5, upper_only=True, value=True)
        malat_val = mad_outlier(adata, 'pct_counts_malat', 5, upper_only=True, value=True)
        log1p_n_genes_val = mad_outlier(adata, 'log1p_n_genes_by_counts', 4, upper_only=False, value=True)
        log1p_total_val = mad_outlier(adata, 'log1p_total_counts', 4, upper_only=False, value=True)
        top_genes_val = mad_outlier(adata, 'pct_counts_in_top_20_genes', 5, upper_only=True, value=True)

        # Create a 2x3 grid of subplots
        fig, axs = plt.subplots(2, 3, figsize=(15, 10))
        axs = axs.flatten()

        violinplot(axs[0], adata, 'pct_counts_mt', mt_val + [25])
        violinplot(axs[1], adata, 'pct_counts_ribo', ribo_val + [30])
        violinplot(axs[2], adata, 'pct_counts_malat', malat_val)
        violinplot(axs[3], adata, 'log1p_n_genes_by_counts', log1p_n_genes_val)
        violinplot(axs[4], adata, 'log1p_total_counts', log1p_total_val)
        violinplot(axs[5], adata, 'pct_counts_in_top_20_genes', top_genes_val)

        plt.tight_layout()

        adata = adata[adata.obs.pct_counts_mt < 25]
        adata = adata[adata.obs.pct_counts_ribo < 30]

        bool_vector = mad_outlier(adata, 'log1p_total_counts', 4) + \
                     mad_outlier(adata, 'log1p_n_genes_by_counts', 4) + \
                     mad_outlier(adata, 'pct_counts_in_top_20_genes', 5, upper_only=True) + \
                     mad_outlier(adata, 'pct_counts_mt', 5, upper_only=True) + \
                     mad_outlier(adata, 'pct_counts_ribo', 5, upper_only=True) + \
                     mad_outlier(adata, 'pct_counts_malat', 5, upper_only=True)

    else:
        mt_val = mad_outlier(adata, 'pct_counts_mt', 5, upper_only=True, value=True)
        ribo_val = mad_outlier(adata, 'pct_counts_ribo', 5, upper_only=True, value=True)
        log1p_n_genes_val = mad_outlier(adata, 'log1p_n_genes_by_counts', 4, upper_only=False, value=True)
        log1p_total_val = mad_outlier(adata, 'log1p_total_counts', 4, upper_only=False, value=True)
        top_genes_val = mad_outlier(adata, 'pct_counts_in_top_20_genes', 5, upper_only=True, value=True)

        # Create a 2x3 grid of subplots (5 plots, one empty)
        fig, axs = plt.subplots(2, 3, figsize=(15, 10))
        axs = axs.flatten()

        violinplot(axs[0], adata, 'pct_counts_mt', mt_val + [5])
        violinplot(axs[1], adata, 'pct_counts_ribo', ribo_val + [10])
        violinplot(axs[2], adata, 'log1p_n_genes_by_counts', log1p_n_genes_val)
        violinplot(axs[3], adata, 'log1p_total_counts', log1p_total_val)
        violinplot(axs[4], adata, 'pct_counts_in_top_20_genes', top_genes_val)
        axs[5].axis('off')  # Turn off the empty subplot

        plt.tight_layout()

        adata = adata[adata.obs.pct_counts_ribo < 10]
        adata = adata[adata.obs.pct_counts_mt < 5]
        bool_vector = mad_outlier(adata, 'log1p_total_counts', 4) + \
                     mad_outlier(adata, 'log1p_n_genes_by_counts', 4) + \
                     mad_outlier(adata, 'pct_counts_in_top_20_genes', 5)

    adata = adata[~bool_vector]
    adata.uns['cells_removed'] = sum(bool_vector)

    return adata, fig

def kneeplot(adata_filtered_final, adata_filtered_init, adata_raw):
    """
    Create a publication-quality knee plot to visualize cell filtering in single-cell RNA sequencing data.

    This function generates a scatter plot showing the relationship between cell rank and
    total UMI counts per cell. The plot follows publication-quality standards suitable for
    journals like Nature, with rank on the x-axis and UMI counts on the y-axis.

    Parameters
    ----------
    adata_filtered_final : AnnData
        Final filtered AnnData object containing selected cells
    adata_filtered_init : AnnData
        Initially filtered AnnData object
    adata_raw : AnnData
        Raw (unfiltered) AnnData object

    Returns
    -------
    tuple
        A tuple containing:
        - matplotlib.figure.Figure: Publication-quality knee plot
        - pandas.DataFrame: DataFrame containing the plotting data and cell classifications

    Notes
    -----
    The plot shows log10-transformed values for both axes:
    - x-axis: log10(rank of cells by total counts)
    - y-axis: log10(total counts per cell)
    """
    filtered_final_obs_names = adata_filtered_final.obs_names
    filtered_init_obs_names = adata_filtered_init.obs_names

    # Ranking barcodes
    adata_raw.obs["sum"] = adata_raw.X.sum(1)

    sum_obs_names_df = pd.DataFrame(adata_raw.obs["sum"])
    sum_obs_names_df.index = adata_raw.obs_names
    # Sort
    sum_obs_names_df = sum_obs_names_df.sort_values(by="sum", ascending=False)  # Changed to descending
    sum_obs_names_df["rank"] = range(1, len(sum_obs_names_df) + 1)  # Start rank from 1
    sum_obs_names_df["log10_sum"] = np.log10(sum_obs_names_df["sum"])
    sum_obs_names_df["log10_rank"] = np.log10(sum_obs_names_df["rank"])
    sum_obs_names_df["filtered_init"] = 0
    sum_obs_names_df.loc[filtered_init_obs_names, "filtered_init"] = 1
    sum_obs_names_df["filtered_final"] = 0
    sum_obs_names_df.loc[filtered_final_obs_names, "filtered_final"] = 1

    # Set the style for publication-quality figure
    plt.style.use('default')
    fig = plt.figure(figsize=(10, 6), dpi=300)

    # Create GridSpec for the main plot
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])

    # Plot with Nature-style formatting
    # Plot black dots for all cells
    ax.scatter(sum_obs_names_df["log10_rank"],
              sum_obs_names_df["log10_sum"],
              color="black",
              s=10,
              alpha=0.3,
              label="All barcodes",
              rasterized=True)  # Rasterize for smaller file size

    # Plot thick light blue dots for initially filtered cells
    init_mask = sum_obs_names_df["filtered_init"] == 1
    ax.scatter(sum_obs_names_df.loc[init_mask, "log10_rank"],
              sum_obs_names_df.loc[init_mask, "log10_sum"],
              color="#4878CF",  # Nature-style blue
              s=100,
              alpha=0.5,
              label="Initial filtering",
              rasterized=True)

    # Plot small red dots for final filtered cells
    final_mask = sum_obs_names_df["filtered_final"] == 1
    ax.scatter(sum_obs_names_df.loc[final_mask, "log10_rank"],
              sum_obs_names_df.loc[final_mask, "log10_sum"],
              color="#E24A33",  # Nature-style red
              s=10,
              alpha=0.85,
              label="Final filtering",
              rasterized=True)

    # Customize the plot for publication quality
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    # Set font sizes
    SMALL_SIZE = 10
    MEDIUM_SIZE = 12
    BIGGER_SIZE = 14

    # Add labels and legend
    ax.set_xlabel("log10(Barcode rank)", fontsize=MEDIUM_SIZE)
    ax.set_ylabel("log10(Total UMI counts)", fontsize=MEDIUM_SIZE)
    ax.tick_params(axis='both', which='major', labelsize=SMALL_SIZE)

    # Add color legend at top right
    leg = ax.legend(frameon=False, fontsize=SMALL_SIZE, markerscale=3,
                   bbox_to_anchor=(1.02, 1), loc='upper right')

    # Adjust layout to prevent label cutoff
    plt.tight_layout()

    # Add gridlines
    ax.grid(True, linestyle='--', alpha=0.3, which='major')

    # Add counts in the corner
    total_cells = len(sum_obs_names_df)
    init_cells = sum_obs_names_df["filtered_init"].sum()
    final_cells = sum_obs_names_df["filtered_final"].sum()

    stats_text = (f'n(total) = {total_cells:,}\n'
                 f'n(initial) = {int(init_cells):,}\n'
                 f'n(final) = {int(final_cells):,}')

    # Add text box with cell counts at bottom left
    plt.text(0.05, 0.05, stats_text,
             transform=ax.transAxes,
             verticalalignment='bottom',
             horizontalalignment='left',
             bbox=dict(boxstyle='round', facecolor='white', alpha=0.8),
             fontsize=SMALL_SIZE)

    return fig, sum_obs_names_df

def create_assignment_barchart(assignments):
    """
    Create a bar chart of cell type assignments and return the figure.

    Parameters:
    -----------
    assignments : list or pandas.Series
        List of cell type assignments

    Returns:
    --------
    matplotlib.figure.Figure
        Figure containing the bar chart
    """
    fig = plt.figure(figsize=(5, 5))
    ax = fig.add_subplot(111)

    pd.Series(assignments).value_counts().plot(kind='bar', stacked=True, ax=ax)

    plt.title('Initial Cellassign Cell Types')
    plt.xlabel('Cell Type')
    plt.ylabel('Number of Cells')
    plt.xticks(rotation=45, ha='right')

    plt.tight_layout()

    return fig

def make_summary_pdf(atlas_df, SRA_ID, fig_kneeplot, violinplots, fig_umap, fig_dotplot, raw_dotplot, fig_barchart,sparsity_original,sparsity,res_sparsity, path="/content/drive/MyDrive/pituitary_atlas"):

    """
    Creates a more compact layout with larger knee plot
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        # Save individual plots
        knee_path = os.path.join(tmpdir, 'knee.png')
        violin_path = os.path.join(tmpdir, 'violin.png')
        bar_path = os.path.join(tmpdir, 'bar.png')
        umap_path = os.path.join(tmpdir, 'umap.png')
        dot_path = os.path.join(tmpdir, 'dot.png')
        raw_dot_path = os.path.join(tmpdir, 'raw_dot.png')

        fig_kneeplot.savefig(knee_path, bbox_inches='tight', dpi=500)
        violinplots.savefig(violin_path, bbox_inches='tight', dpi=500)
        fig_barchart.savefig(bar_path, bbox_inches='tight', dpi=500)
        fig_umap.savefig(umap_path, bbox_inches='tight', dpi=500)
        fig_dotplot.savefig(dot_path, bbox_inches='tight', dpi=400)
        raw_dotplot.savefig(raw_dot_path, bbox_inches='tight', dpi=400)

        # Create summary figure - reduced height
        fig = plt.figure(figsize=(20, 24))
        # 4 rows, reduced height ratios and spacing
        gs = GridSpec(5, 5, height_ratios=[0.25, 0.35, 0.35, 0.4,0.4], figure=fig)
        gs.update(wspace=0.04, hspace=0.08)  # Reduced spacing further

        # Metadata
        meta_ax = fig.add_subplot(gs[0, :])
        meta_ax.axis('off')
        meta_row = atlas_df[atlas_df['SRA_ID'] == SRA_ID].iloc[0]
        metadata_text = (
            f"Sample Information:\n"
            f"==================\n"
            f"Author: {meta_row['Author']}\n"
            f"SRA_ID: {meta_row['SRA_ID']}\n"
            f"GEO: {meta_row['GEO']}\n"
            f"Age_numeric: {meta_row['Age_numeric']}\n"
            f"Name: {meta_row['Name']}\n"
            f"Comp_sex: {meta_row['Comp_sex']}\n"
            f"Number of cells: {meta_row['n_cells']}\n"
            f"Initial sparsity: {sparsity_original}\n"
            f"Resulting sparsity: {res_sparsity}\n"
            f"Median CellAssign probability: {meta_row['median_cellassign_prob']:.3f}"
        )
        meta_ax.text(0.02, 0.95, metadata_text,
                    fontsize=10,
                    verticalalignment='top',
                    fontfamily='monospace',
                    bbox=dict(facecolor='white',
                             edgecolor='black',
                             alpha=0.8))

        # Violin plot and Knee plot row
        violin_ax = fig.add_subplot(gs[1, :3])  # Violin takes 3/5 of width
        violin_ax.imshow(plt.imread(violin_path))
        violin_ax.axis('off')

        knee_ax = fig.add_subplot(gs[1, 3:])  # Knee takes 2/5 of width
        knee_ax.imshow(plt.imread(knee_path))
        knee_ax.axis('off')

        # UMAP and Bar plot row
        umap_ax = fig.add_subplot(gs[2, :3])  # UMAP takes 3/5 of width
        umap_ax.imshow(plt.imread(umap_path))
        umap_ax.axis('off')

        bar_ax = fig.add_subplot(gs[2, 3:])  # Bar takes 2/5 of width
        bar_ax.imshow(plt.imread(bar_path))
        bar_ax.axis('off')

        #raw dotplot

        raw_dot_ax = fig.add_subplot(gs[3, :])
        raw_dot_ax.imshow(plt.imread(raw_dot_path))
        raw_dot_ax.axis('off')

        # Dot plot
        dot_ax = fig.add_subplot(gs[4, :])
        dot_ax.imshow(plt.imread(dot_path))
        dot_ax.axis('off')

        # Save final figure
        plt.savefig(f"{path}/processed/nac/{SRA_ID}/analysis/summary.pdf",
                    bbox_inches='tight', dpi=400, pad_inches=0)  # Reduced padding
        plt.close('all')

def get_sparsity(adata):
  if sp.issparse(adata.X):
    zero_entries = (adata.X == 0).sum()
    non_zero_entries = (adata.X != 0).sum()
    sparsity = zero_entries / (non_zero_entries + zero_entries)
  else:
    zero_entries = np.sum(adata.X == 0)
    non_zero_entries = np.sum(adata.X != 0)
    sparsity = zero_entries / (non_zero_entries + zero_entries)
  return sparsity

def shrink_high_values_normalized_df(df, column_name="ambient_profile_mrna", percentile=50, shrinkage_factor=0.5):
    """
    Shrink values above the specified percentile toward that percentile value and normalize
    the column so it sums to 1, while keeping zero values and lower values unchanged.

    Args:
        df (pd.DataFrame): A Pandas DataFrame containing the data.
        column_name (str): The name of the column to process.
        percentile (float): The percentile threshold (default 80).
        shrinkage_factor (float): Factor determining how much to shrink (0-1, default 0.5).

    Returns:
        pd.DataFrame: A new DataFrame with the specified column adjusted and normalized.
    """
    # Create a copy of the DataFrame
    df_copy = df.copy()

    # Extract the column to process
    data = df_copy[column_name]

    # Calculate the percentile threshold using only non-zero values
    non_zero_data = data[data != 0]
    percentile_value = np.percentile(non_zero_data, percentile)

    # Function to shrink values above the threshold
    def shrink_high_value(x):
        if x > percentile_value:
            # Shrink towards the percentile value
            return x + (percentile_value - x) * shrinkage_factor
        return x

    # Apply shrinkage only to values above the threshold
    adjusted_data = data.copy()
    # Only process non-zero values
    adjusted_data[data != 0] = non_zero_data.apply(shrink_high_value)

    # Normalize to sum to 1
    total_sum = adjusted_data.sum()
    if total_sum > 0:
        adjusted_data = adjusted_data / total_sum

    # Update the DataFrame copy
    df_copy[column_name] = adjusted_data

    return df_copy


### Code for the ATAC processing
def qc_atac(adata, log_n_frag_mad=3):
    #
    adata.obs["log_n_frags"] = np.log2(adata.obs['n_fragment'] )
    #keep top 10K cells
    data_array = np.array(adata.obs["n_fragments"])
    sorted_indices = np.argsort(data_array)[::-1]
    keep = sorted_indices[:10000].tolist()
    
    adata = adata[keep].copy()

    log_n_frags = mad_outlier(adata, "log_n_frags", log_n_frag_mad, upper_only=False, value=True)

    # Create a 2x3 grid of subplots (5 plots, one empty)
    fig, axs = plt.subplots(1, 2, figsize=(7, 6))
    axs = axs.flatten()

    violinplot(axs[0], adata, "log_n_frags", log_n_frags + [10.55])
    violinplot(axs[1], adata, 'tsse', [7])
    plt.tight_layout()

    keep = np.array(adata.obs_names)[~mad_outlier(adata, "log_n_frags", 3).to_numpy() ]

    return adata, fig


import numpy as np
import scipy.sparse as sp
import joblib
from pathlib import Path
import xgboost as xgb
from .celltyping import load_celltype_model, prepare_matrix_celltype, perform_celltype_prediction,smoothing_cell_types
from .doublets import load_doublet_model, prepare_matrix_doublet, perform_doublet_prediction

def check_sample_compatibility_features(adata, feature_names, return_present=True,
                                        return_missing=True):
    varnames = adata.var_names

    #check percentage of features that are present in the adata object
    present = np.isin(feature_names, varnames)
    present_percentage = np.sum(present) / len(feature_names) * 100
    print(f"Percentage of features present in adata: {present_percentage:.2f}%")
    if return_present:
        present_features = feature_names[present]
        print(f"Number of features present in adata: {len(present_features)}")
    else:
        present_features = None
    if return_missing:
        missing_features = feature_names[~present]
        print(f"Number of features missing in adata: {len(missing_features)}")
    else:
        missing_features = None

    #pass if at least 70% of the features are present
    if present_percentage < 70:
        print("Warning: Less than 70% of the features are present in the adata object.")
        passing = False
    else:
        print("At least 70% of the features are present in the adata object.")
        passing = True

    
    if return_present and return_missing:
        return passing, present_features, missing_features
    
    elif return_present:
        return passing, present_features
    elif return_missing:
        return passing, missing_features
    else:
        return passing
    

def check_sample_compatibility_normalization(adata, force=False):
    # Check if the model is for RNA or ATAC
    #check number of cells, if more than 50000, throw an error
    if adata.shape[0] > 50000:
        print(f"Error: The number of cells in the dataset is greater than 50000. Are you sure this has been filtered correctly?")
    
    #take info from first 10 cells
    first_20_cells = adata[:20, :].X
    #check if they have integer values or if 1.0 occurs more than 10% of the time
    integer_type = np.issubdtype(first_20_cells.dtype, np.integer)
    if integer_type:
        print(f"Warning: The dataset appears to be in integer format. Are you sure this has been normalized correctly?")
    
    not_normed = False
    one_values = np.sum(first_20_cells == 1.0) / first_20_cells.size
    if one_values > 0.05:
        not_normed = True
        print(f"Warning: The dataset has more than 5% of the values equal to 1.0. Are you sure this has been normalized correctly?")

    one_values = np.sum(first_20_cells == 1) / first_20_cells.size
    if one_values > 0.05:
        not_normed = True
        print(f"Warning: The dataset has more than 5% of the values equal to 1.0. Are you sure this has been normalized correctly?")

    #if all cells sum to nearly 10k, say it hasnt been logged
    not_logged = False
    if np.all(np.sum(first_20_cells, axis=1) > 9500) and np.all(np.sum(first_20_cells, axis=1) < 10500):
        print(f"Warning: Have you logged the dataset? The cells sum to nearly 10k.")
        not_logged = True
    
    #if force is True, return True
    passing = False
    if force:
        print(f"Warning: Force is set to True. Passing the dataset compatibility check.")
        passing = True
    elif not integer_type and one_values < 0.1 and not not_logged:
        print(f"The dataset has passed the compatibility check.")
        passing = True

    return passing, not_normed, not_logged


def pct_counts_kept(adata):
  counts_per_10k_current = adata.X.sum(axis=1) / 10000

  return np.asarray(counts_per_10k_current).flatten()


def calc_pct_counts_kept(adata, features):
    copied_adata = adata.copy()
    features_in_data = [f for f in features if f in copied_adata.var_names]
    copied_adata = copied_adata[:, features_in_data].copy()
    #undo log1p
    if sp.issparse(copied_adata.X):
        copied_adata.X = copied_adata.X.expm1()
    else:
        copied_adata.X = np.expm1(copied_adata.X)

    adata.obs["pct_counts_kept"] = pct_counts_kept(copied_adata)
    return adata


def get_base_path():
    """Get the absolute path to the project root directory."""
    # Check for environment variable first
    return Path(__file__).parent


def cell_type_workflow(adata_to_use, active_assay="sc",modality="rna",in_place=True, nan_or_zero='nan',smoothing=True):
    """
    Main workflow for cell type prediction.
    """
    if in_place:
        adata = adata_to_use
    else:
        adata = adata_to_use.copy()
    
    adata.var_names_make_unique()

    base_path = get_base_path()
    if modality == "rna":
        model_path = f"{base_path}/models/rna_model.json"
        label_encoder_path = f'{base_path}/models/label_encoder_rna.pkl'

    elif modality == "atac":
        model_path = f"/{base_path}/models/atac_model.json"
        label_encoder_path = f'{base_path}/models/label_encoder_atac.pkl'


    model, label_encoder, feature_names = load_celltype_model(model_path, label_encoder_path)

    #checks
    check_sample_compatibility_features(adata, feature_names, return_present=False, return_missing=False)
    check_sample_compatibility_normalization(adata, force=False)

    # Prepare the matrix for cell type prediction
    X_final = prepare_matrix_celltype(adata, feature_names, active_assay=active_assay, nan_or_zero=nan_or_zero)
    # Perform cell type prediction
    predicted_cell_types, probas = perform_celltype_prediction(X_final, model, label_encoder)
    # Add predictions to adata
    
    adata.obs['predicted_cell_type'] = predicted_cell_types
    adata.obs['predicted_cell_type_proba'] = probas.max(axis=1)  # Store max probability

    for i, label in enumerate(label_encoder.classes_):
        adata.obs[f'proba_{label}'] = probas[:, i]


    if smoothing:
        adata = smoothing_cell_types(adata)
    return adata


def doublet_workflow(adata_to_use,active_assay="sc",modality="rna",in_place=True, nan_or_zero='nan'):

    if in_place:
        adata = adata_to_use
    else:
        adata = adata_to_use.copy()
    adata.var_names_make_unique()

    base_path = get_base_path()

    if modality == "rna":
        model_path = f"{base_path}/models/rna_model_binary.json"
        label_encoder_path = f'{base_path}/models/rna_label_encoder_binary.pkl'
        threshold_path = f'{base_path}/models/final_threshold.pkl'

    elif modality == "atac":
        model_path = f"{base_path}/models/atac_model_binary.json"
        label_encoder_path = f'{base_path}/models/atac_label_encoder_binary.pkl'
        threshold_path = f'{base_path}/models/atac_final_threshold.pkl'

    model, label_encoder, threshold, feature_names = load_doublet_model(model_path, label_encoder_path, threshold_path)


    #checks
    check_sample_compatibility_features(adata, feature_names, return_present=False, return_missing=False)
    check_sample_compatibility_normalization(adata, force=False)

    adata = calc_pct_counts_kept(adata, feature_names)
    # Prepare the matrix for doublet prediction
    X_final = prepare_matrix_doublet(adata, feature_names, active_assay=active_assay, nan_or_zero=nan_or_zero)
    # Perform doublet prediction
    predicted_doublet_labels, is_doublet, doublet_score = perform_doublet_prediction(X_final, model, label_encoder, threshold)
    # Add predictions to adata
    
    adata.obs['init_predicted_doublet_epitome'] = predicted_doublet_labels
    adata.obs['thresholded_doublet_epitome'] = is_doublet
    adata.obs['doublet_score_epitome'] = doublet_score
    return adata

def celltype_doublet_workflow(adata, active_assay="sc", modality="rna", in_place=True, nan_or_zero='nan', smoothing=True):
    """
    Main workflow for cell type and doublet prediction.
    """
    adata = cell_type_workflow(adata, active_assay=active_assay, modality=modality, in_place=in_place, nan_or_zero=nan_or_zero, smoothing=smoothing)
    adata = doublet_workflow(adata, active_assay=active_assay, modality=modality, in_place=in_place, nan_or_zero=nan_or_zero)
    return adata


from setuptools import setup, find_packages

VERSION = '1.0.0'
DESCRIPTION = 'Auxiliary tools for the Consensus Pituitary Atlas'
LONG_DESCRIPTION = 'Python package containing auxiliary tools for the Consensus Pituitary Atlas. The current workflow commands allow celltyping and doublet detection with a single line of code, for more information see https://github.com/BKover99/epitome_tools'


setup(
    name="epitome_tools",
    version=VERSION,
    author="Bence Kover",
    author_email="kover.bence@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/plain",
    packages=find_packages(),
package_data={
        'epitome_tools': ['models/*'],
    },
    install_requires=[
	'xgboost',
	'scipy',
	'numpy'
	],
include_package_data=True,
    keywords=['xgboost', 'annotation', 'celltype', 'doublet'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X"
    ]
)




# __init__.py
from .celltyping import load_celltype_model, prepare_matrix_celltype, perform_celltype_prediction
from .doublets import load_doublet_model, prepare_matrix_doublet, perform_doublet_prediction
from .workflow import celltype_doublet_workflow, doublet_workflow, cell_type_workflow

__all__ = [
    'load_celltype_model', 'prepare_matrix_celltype', 'perform_celltype_prediction',
    'load_doublet_model', 'prepare_matrix_doublet', 'perform_doublet_prediction',
    'celltype_doublet_workflow', 'doublet_workflow', 'cell_type_workflow'
]